#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import base64
from oslo_config import cfg
from oslo_log import log as logging
from oslo_middleware import base
import requests
from tacker.api.vnflcm.v1 import router as vnflcm_router
from tacker.api.vnfpkgm.v1 import router as vnfpkgm_router
import threading
import webob.dec
import webob.exc

from tacker.common import utils
from tacker import context

LOG = logging.getLogger(__name__)


class TackerKeystoneContext(base.ConfigurableMiddleware):
    """Make a request context from keystone headers."""

    @webob.dec.wsgify
    def __call__(self, req):
        ctx = context.Context.from_environ(req.environ)

        if not ctx.user_id:
            LOG.debug("X_USER_ID is not found in request")
            return webob.exc.HTTPUnauthorized()

        # Inject the context...
        req.environ['tacker.context'] = ctx

        return self.application


class _BearerAuth(requests.auth.AuthBase):
    """Attaches HTTP Bearer Authentication to the given Request object."""

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = "Bearer " + self.token
        return r


class _OAuth2GrantBase(metaclass=abc.ABCMeta):
    """Base class that all OAuth2.0 grant type implementations derive from."""

    grant_type = None

    @abc.abstractmethod
    def get_accsess_token(self):
        pass


class _ClientCredentialsGrant(_OAuth2GrantBase):
    """OAuth2.0 grant type ClientCredentials implementation."""

    grant_type = 'client_credentials'

    def __init__(self, token_endpoint, client_id, client_password):
        super(_ClientCredentialsGrant, self).__init__()
        self.token_endpoint = token_endpoint
        self.client_id = client_id
        self.client_password = client_password

    def get_accsess_token(self):
        """Get access token.

        Returns:
            dict: access token response.
        """
        kwargs = {
            'headers': {
                'Connection': 'keep-alive',
                'Content-Type': 'application/x-www-form-urlencoded'},
            'data': {
                'grant_type': self.grant_type},
            'timeout': cfg.CONF.authentication.timeout}

        basic_auth_request = _BasicAuthSession(
            self.client_id, self.client_password)

        LOG.info(
            "Get Access Token, Connecting to <GET:{}>".format(
                self.token_endpoint))
        LOG.info("Request Headers={}".format(kwargs.get('headers')))
        LOG.info("Request Body={}".format(kwargs.get('data')))

        response = basic_auth_request.post(
            self.token_endpoint,
            verify=cfg.CONF.authentication.verify_oauth2_ssl,
            **kwargs)
        response.raise_for_status()

        response_body = response.json()
        LOG.info("[RES] Headers={}".format(response.headers))
        LOG.info("[RES] Body={}".format(response_body))

        return response_body


class _OAuth2Session(requests.Session):
    """Provides OAuth 2.0 authentication."""

    def __init__(self, grant):
        super(_OAuth2Session, self).__init__()
        self.grant = grant
        self.__access_token_info = {}
        self.__lock = threading.RLock()

    def request(self, method, url, **kwargs):
        """Override <requests.Session.request> function."""
        kwargs['auth'] = _BearerAuth(
            self.__access_token_info.get('access_token'))

        response = super().request(method, url, **kwargs)
        if response.status_code == 401:
            LOG.error(
                'Authentication error {}, details={}'.format(
                    response, response.text))
            self.apply_access_token_info()

        return response

    def apply_access_token_info(self):
        """Get access token."""
        try:
            self.__set_access_token_info(self.grant.get_accsess_token())
            self.schedule_refrash_token()
        except requests.exceptions.RequestException as e:
            if hasattr(e, 'response') and e.response:
                LOG.error(
                    "Get Access Token, error details={}".format(
                        e.response.json()))
            LOG.error(e)

    def __set_access_token_info(self, update_dict):
        with self.__lock:
            self.__access_token_info = update_dict

    def schedule_refrash_token(self):
        """expires_in Scheduler registration at expiration."""
        if not ('expires_in' in self.__access_token_info):
            LOG.debug("'expires_in' does not exist in the response body.")
            return

        try:
            expires_in = int(self.__access_token_info.get('expires_in'))
            expires_in_timer = threading.Timer(
                expires_in, self.apply_access_token_info)
            expires_in_timer.start()

            LOG.info(
                "expires_in=<{}> exist, scheduler regist.".format(expires_in))
        except (ValueError, TypeError):
            pass


class _BasicAuthSession(requests.Session):
    """Provide Basic authentication."""

    def __init__(self, user_name, password):
        super(_BasicAuthSession, self).__init__()
        self.user_name = user_name
        self.password = password
        self.auth = requests.auth.HTTPBasicAuth(
            self.user_name, self.password)

    def request(self, method, url, **kwargs):
        """Override <requests.Session.request> function."""
        kwargs['auth'] = self.auth
        return super().request(method, url, **kwargs)


class _AuthManager:

    OPTS = [
        cfg.StrOpt('auth_type',
                default=None,
                choices=['BASIC', 'OAUTH2_CLIENT_CREDENTIALS'],
                help="auth_type used for external connection"),
        cfg.IntOpt('timeout',
                default=20,
                help="timeout used for external connection"),
        cfg.StrOpt('token_endpoint',
                default=None,
                help="token_endpoint used to get the oauth2 token"),
        cfg.StrOpt('client_id',
                   default=None,
                   help="client_id used to get the oauth2 token"),
        cfg.StrOpt('client_password',
                default=None,
                help="client_password used to get the oauth2 token"),
        cfg.StrOpt('user_name',
                default=None,
                help="user_name used in basic authentication"),
        cfg.StrOpt('password',
                default=None,
                help="password used in basic authentication"),
        cfg.BoolOpt('verify_oauth2_ssl',
                default=True,
                help="verify the certification to get the oauth2 access token "
                     "by ssl")
    ]
    cfg.CONF.register_opts(OPTS, group='authentication')

    __DEFAULT_CLIENT = requests.Session()

    def __init__(self):
        self.__manages = {}
        self.__lock = threading.RLock()

        # local auth setting.
        self.set_auth_client(
            auth_type=cfg.CONF.authentication.auth_type,
            auth_params={'client_id': cfg.CONF.authentication.client_id,
            'client_password': cfg.CONF.authentication.client_password,
            'token_endpoint': cfg.CONF.authentication.token_endpoint,
            'user_name': cfg.CONF.authentication.user_name,
            'password': cfg.CONF.authentication.password})

    def __empty(self, val):
        if val is None:
            return True
        elif isinstance(val, str):
            return val.strip() == ''

        return len(val) == 0

    def set_auth_client(self, id='local', auth_type=None, auth_params=None):
        """Set up an Auth client.

        Args:
            id (str, optional): Management ID
            auth_type (str, optional): Authentication type.
            auth_params (dict, optional): Authentication information.
        """
        snakecase_auth_params = utils.convert_camelcase_to_snakecase(
            auth_params)
        if self.__empty(auth_type) or self.__empty(snakecase_auth_params):
            return

        if id in self.__manages:
            LOG.debug("Use cache, Auth Managed Id=<{}>".format(id))
            return

        client = self.__DEFAULT_CLIENT
        if auth_type == 'BASIC':
            client = _BasicAuthSession(
                user_name=snakecase_auth_params.get('user_name'),
                password=snakecase_auth_params.get('password'))
        elif (auth_type == 'OAUTH2_CLIENT_CREDENTIALS' and
                not self.__empty(snakecase_auth_params.get('token_endpoint'))):
            grant = _ClientCredentialsGrant(
                client_id=snakecase_auth_params.get('client_id'),
                client_password=snakecase_auth_params.get('client_password'),
                token_endpoint=snakecase_auth_params.get('token_endpoint'))
            client = _OAuth2Session(grant)
            client.apply_access_token_info()

        LOG.info(
            "Add to Auth management, id=<{}>, type=<{}>, class=<{}>".format(
                id, auth_type, client.__class__.__name__))

        self.__add_manages(id, client)

    def __add_manages(self, id, client):
        with self.__lock:
            self.__manages[id] = client

    def get_auth_client(self, id="local"):
        """Get the Auth client.

        Args:
            id (str, optional): Management ID

        Returns:
            based on <requests.Session> class.
        """
        return self.__manages.get(id, self.__DEFAULT_CLIENT)


class _AuthBase(metaclass=abc.ABCMeta):

    def __init__(self, api_name, token_type, token_value):
        self.api_name = api_name
        self.token_type = token_type
        self.token_value = token_value

    @abc.abstractmethod
    def do_auth(self):
        pass

    def _check_token_type(self):
        if self.token_type == cfg.CONF.authentication.token_type:
            return

        msg = ("Auth Scheme is not expected token_type: expect={%s},\
        actual={%s}" % (cfg.CONF.authentication.token_type, self.token_type))
        raise webob.exc.HTTPUnauthorized(msg)


class _AuthValidateIgnore(_AuthBase):
    def __init__(self, api_name, token_type, token_value):
        super().__init__(api_name, token_type, token_value)

    def do_auth(self):
        conf_auth_type = cfg.CONF.authentication.token_type
        if (conf_auth_type is not None and self.token_type is None):
            msg = ("Request header is None but tacker conf\
                has auth information.")
            raise webob.exc.HTTPUnauthorized(msg)

        return None


class _AuthValidateBearer(_AuthBase):

    def __init__(self, application_type, api_name, token_type, token_value):
        super().__init__(api_name, token_type, token_value)
        self.application_type = application_type
        self.expires_in = 0
        self.res_access_token = None
        self.__access_token_info = {}

    def request(self):
        auth_url = cfg.CONF.authentication.auth_url
        response = requests.Session().get(auth_url)
        if response.status_code == 401:
            return None

        return response

    def do_auth(self):
        self._check_token_type()

        if self.res_access_token is None:

            response = self.request()

            if response is None:
                msg = "No responce from Authorization Server."
                raise webob.exc.HTTPUnauthorized(msg)

            response_body = response.json()

            if (response_body.get('access_token') is None or
                    response_body.get('token_type') is None):
                msg = "No access_token or token_type exist."
                raise webob.exc.HTTPUnauthorized(msg)

            if self.token_value != response_body.get('access_token'):
                msg = "access_token is invalid."
                raise webob.exc.HTTPUnauthorized(msg)

            self.expires_in = response_body.get('expires_in')
            self.res_access_token = response_body.get('access_token')

            self._validate_scope(response_body.get('scope'))

            self._scheduler_delete_token(response_body)

    def _scheduler_delete_token(self, response_body):
        if not (response_body.get('expires_in')):
            LOG.debug("'expires_in' does not exist in the response body.")
            return

        try:
            expires_in = int(response_body.get('expires_in'))
            expires_in_timer = threading.Timer(
                expires_in, self._delete_access_token)
            expires_in_timer.start()

            LOG.info(
                "expires_in=<{}> exist, scheduler regist.".format(expires_in))
        except (ValueError, TypeError):
            pass

    def _delete_access_token(self):
        self.access_token = None
        self.expires_in = 0

    def _generate_api_scope_name(self):
        if self.application_type == vnflcm_router.VnflcmAPIRouter:
            scope_prefix = 'vnflcm_'
            return scope_prefix + self.api_name
        elif self.application_type == vnfpkgm_router.VnfpkgmAPIRouter:
            scope_prefix = 'vnfpkgm_'
            return scope_prefix + self.api_name

        return ''

    def _validate_scope(self, res_scope):
        if res_scope is None:
            return

        try:
            api_scope_name = self._generate_api_scope_name() + '_scope'
            scopes = cfg.CONF.authentication.__getitem__(api_scope_name)
        except Exception:
            msg = ("Getting scope error."
                  "api_scope_name={%s}, scopes={%s}",
                  api_scope_name, scopes)
            raise webob.exc.HTTPUnauthorized(msg)

        if len(scopes) == 0:
            return

        if res_scope in scopes:
            return

        raise webob.exc.HTTPForbidden("scope is undefined in tacker.conf")


class _AuthValidateBasic(_AuthBase):

    def __init__(self, api_name, token_type, token_value):
        super().__init__(api_name, token_type, token_value)

    def do_auth(self):
        self._check_token_type()

        base = cfg.CONF.authentication.user_name +\
            cfg.CONF.authentication.password
        base64_encode = base64.b64encode(base.encode())
        if self.token_value != base64_encode:
            msg = "access_token and encoded base64 are not same."
            raise webob.exc.HTTPUnauthorized(msg)


class _AuthValidateManager:

    atuh_opts = [
        cfg.StrOpt('token_type',
                default=None,
                choices=['Bearer', 'Basic'],
                help="authentication type"),
        cfg.StrOpt('user_name',
                default=None,
                help="user_name used in basic authentication"),
        cfg.StrOpt('password',
                default=None,
                help="password used in basic authentication"),
        cfg.StrOpt('auth_url',
                default=None,
                help="URL of the authorization server")
    ]
    cfg.CONF.register_opts(atuh_opts, group='authentication')

    def __init__(self):
        self.__manages = {}
        self.__lock = threading.RLock()

    def __add_manages(self, token_value, auth_obj):
        with self.__lock:
            self.__manages[token_value] = auth_obj

    def _parse_request_header(self, request):
        auth_req = request.headers.get('Authorization')
        if auth_req is None:
            return (None, None)

        auth_req_arry = auth_req.split(' ')
        if len(auth_req_arry) > 3:
            msg = "Invalid Authorization header."
            raise webob.exc.HTTPUnauthorized(msg)
        return (auth_req_arry[0], auth_req_arry[1])

    def _get_auth_type(self, request, application):
        token_type, token_value = self._parse_request_header(request)

        if token_value in self.__manages:
            return self.__manages.get(token_value)

        match = application.map.match(request.path_info)
        api_name = match[0].get("action")

        if token_type == 'Bearer':
            auth_obj = _AuthValidateBearer(
                type(application), api_name, token_type, token_value)
            self.__add_manages(token_value, auth_obj)
            return auth_obj
        elif token_type == 'Basic':
            auth_obj = _AuthValidateBasic(api_name, token_type, token_value)
            self.__add_manages(token_value, auth_obj)
            return auth_obj

        return _AuthValidateIgnore(api_name, token_type, token_value)

    def auth_main(self, request, application):
        auth_validator = self._get_auth_type(request, application)
        if auth_validator:
            auth_validator.do_auth()


class AuthValidatorExecution(base.ConfigurableMiddleware):

    @ webob.dec.wsgify
    def __call__(self, req):
        auth_validator_manager.auth_main(req, self.application)
        return self.application


def pipeline_factory(loader, global_conf, **local_conf):
    """Create a paste pipeline based on the 'auth_strategy' config option."""
    pipeline = local_conf[cfg.CONF.auth_strategy]
    pipeline = pipeline.split()
    filters = [loader.get_filter(n) for n in pipeline[:-1]]
    app = loader.get_app(pipeline[-1])
    filters.reverse()
    for f in filters:
        app = f(app)
    return app


auth_manager = _AuthManager()
auth_validator_manager = _AuthValidateManager()



# Copyright (C) 2021 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from tacker.common.rpc import BackingOffClient
from tacker.conductor.conductorrpc import vnf_lcm_rpc
from tacker.tests import base


class VnfLcmRPCTestCase(base.BaseTestCase):

    def setUp(self):
        super(VnfLcmRPCTestCase, self).setUp()
        self.context = self.fake_admin_context()
        self.rpc_api = vnf_lcm_rpc.VNFLcmRPCAPI()
        self.cctxt_mock = mock.MagicMock()

    @mock.patch.object(BackingOffClient, 'prepare')
    def test_instantiate(self, prepare_mock):
        vnf_instance = ""
        vnf = ""
        instantiate_vnf = ""
        vnf_lcm_op_occs_id = ""
        prepare_mock.return_value = self.cctxt_mock
        self.rpc_api.instantiate(
            self.context, vnf_instance,
            vnf, instantiate_vnf, vnf_lcm_op_occs_id)
        prepare_mock.assert_called()
        self.cctxt_mock.cast.assert_called_once_with(
            self.context, 'instantiate',
            vnf_instance=vnf_instance,
            vnf_dict=vnf,
            instantiate_vnf=instantiate_vnf,
            vnf_lcm_op_occs_id=vnf_lcm_op_occs_id)

    @mock.patch.object(BackingOffClient, 'prepare')
    def test_terminate(self, prepare_mock):
        vnf_instance = ""
        vnf = ""
        terminate_vnf_req = ""
        vnf_lcm_op_occs_id = ""
        prepare_mock.return_value = self.cctxt_mock
        self.rpc_api.terminate(
            self.context, terminate_vnf_req,
            vnf, terminate_vnf_req, vnf_lcm_op_occs_id)
        prepare_mock.assert_called()
        self.cctxt_mock.cast.assert_called_once_with(
            self.context, 'terminate',
            vnf_instance=vnf_instance,
            vnf_dict=vnf,
            terminate_vnf_req=terminate_vnf_req,
            vnf_lcm_op_occs_id=vnf_lcm_op_occs_id)

    @mock.patch.object(BackingOffClient, 'prepare')
    def test_heal(self, prepare_mock):
        vnf_instance = ""
        vnf = ""
        heal_vnf_request = ""
        vnf_lcm_op_occs_id = ""
        prepare_mock.return_value = self.cctxt_mock
        self.rpc_api.heal(
            self.context, heal_vnf_request,
            vnf, heal_vnf_request, vnf_lcm_op_occs_id)
        prepare_mock.assert_called()
        self.cctxt_mock.cast.assert_called_once_with(
            self.context, 'heal',
            vnf_instance=vnf_instance,
            vnf_dict=vnf,
            heal_vnf_request=heal_vnf_request,
            vnf_lcm_op_occs_id=vnf_lcm_op_occs_id)

    @mock.patch.object(BackingOffClient, 'prepare')
    def test_scale(self, prepare_mock):
        vnf_info = ""
        vnf_instance = ""
        scale_vnf_request = ""
        prepare_mock.return_value = self.cctxt_mock
        self.rpc_api.scale(
            self.context, vnf_info,
            vnf_instance, scale_vnf_request)
        prepare_mock.assert_called()
        self.cctxt_mock.cast.assert_called_once_with(
            self.context, 'scale',
            vnf_info=vnf_info,
            vnf_instance=vnf_instance,
            scale_vnf_request=scale_vnf_request)

    @mock.patch.object(BackingOffClient, 'prepare')
    def test_update(self, prepare_mock):
        vnf_lcm_opoccs = ""
        body_data = ""
        vnfd_pkg_data = ""
        vnfd_id = ""
        prepare_mock.return_value = self.cctxt_mock
        self.rpc_api.update(
            self.context, vnf_lcm_opoccs,
            body_data, vnfd_pkg_data, vnfd_id)
        prepare_mock.assert_called()
        self.cctxt_mock.cast.assert_called_once_with(
            self.context, 'update',
            vnf_lcm_opoccs=vnf_lcm_opoccs,
            body_data=body_data,
            vnfd_pkg_data=vnfd_pkg_data,
            vnfd_id=vnfd_id)

    @mock.patch.object(BackingOffClient, 'prepare')
    def test_rollback(self, prepare_mock):
        vnf_info = ""
        vnf_instance = ""
        operation_params = ""
        prepare_mock.return_value = self.cctxt_mock
        self.rpc_api.rollback(
            self.context, vnf_info,
            vnf_instance, operation_params)
        prepare_mock.assert_called()
        self.cctxt_mock.cast.assert_called_once_with(
            self.context, 'rollback',
            vnf_info=vnf_info,
            vnf_instance=vnf_instance,
            operation_params=operation_params)



# Copyright (C) 2022 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Setup privsep decorator."""

from oslo_privsep import capabilities as caps
from oslo_privsep import priv_context

default = priv_context.PrivContext(
    __name__,
    cfg_section='privsep',
    pypath="f{__name__}.default",
    capabilities=[caps.CAP_SYS_ADMIN,
                  caps.CAP_NET_ADMIN,
                  caps.CAP_DAC_OVERRIDE,
                  caps.CAP_DAC_READ_SEARCH,
                  caps.CAP_SYS_PTRACE],
)



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import unittest

from tacker.tests.functional.sol_kubernetes.vnflcm import base as vnflcm_base


class VnfLcmKubernetesScaleTest(vnflcm_base.BaseVnfLcmKubernetesTest):

    @classmethod
    def setUpClass(cls):
        super(VnfLcmKubernetesScaleTest, cls).setUpClass()
        vnf_package_id, cls.vnfd_id = \
            cls._create_and_upload_vnf_package(
                cls, cls.tacker_client, "test_cnf_scale",
                {"key": "sample_scale_functional"})
        cls.vnf_package_ids.append(vnf_package_id)

    @classmethod
    def tearDownClass(cls):
        super(VnfLcmKubernetesScaleTest, cls).tearDownClass()

    def _test_cnf_scale(self, vnf_instance, aspect_id,
                        number_of_steps=1, error=False):
        scale_level = self._get_scale_level_by_aspect_id(
            vnf_instance, aspect_id)

        # test scale out
        scale_level = self._test_scale(
            vnf_instance['id'], 'SCALE_OUT', aspect_id, scale_level,
            number_of_steps, error)
        if error:
            return scale_level

        # test scale in
        scale_level = self._test_scale(
            vnf_instance['id'], 'SCALE_IN', aspect_id, scale_level,
            number_of_steps)

        return scale_level

    def test_scale_cnf_with_statefulset(self):
        """Test scale for CNF (StatefulSet)

        This test will instantiate cnf with StatefulSet and scale replicas.
        """
        vnf_instance_name = "cnf_scale_with_statefulset"
        vnf_instance_description = "cnf scale with statefulset"
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/statefulset_scale.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "simple", vnf_instance_name,
            vnf_instance_description, inst_additional_param)
        self._test_cnf_scale(vnf_instance, "vdu1_aspect")
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])

    def test_scale_cnf_with_replicaset(self):
        """Test scale for CNF (ReplicaSet)

        This test will instantiate cnf with ReplicaSet and scale replicas.
        """
        vnf_instance_name = "cnf_scale_with_replicaset"
        vnf_instance_description = "cnf scale with replicaset"
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/replicaset_scale.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "simple", vnf_instance_name,
            vnf_instance_description, inst_additional_param)
        self._test_cnf_scale(vnf_instance, "vdu1_aspect")
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])

    def test_scale_cnf_deployment_with_scaling_and_delta_two(self):
        """Test scale for CNF (Deployment)

        This test will instantiate cnf with Deployment and scale replicas.
        And scaling steps of ScaleVnfRequest set to two and scaling deltas that
        defined in VNFD set to two.
        """
        vnf_instance_name = "cnf_scale_with_scaling_and_delta_two"
        vnf_instance_description = "cnf scale with scaling and delta two"
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/deployment_scale.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "scalingsteps", vnf_instance_name,
            vnf_instance_description, inst_additional_param)
        # Use flavour_id scalingsteps that is set to delta_num=2
        self._test_cnf_scale(vnf_instance, "vdu1_aspect", number_of_steps=2)
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])

    @unittest.skip("Reduce test time")
    def test_scale_out_cnf_rollback(self):
        """Test rollback after scaling failure for CNF

        This test will rollback after failing scale out operation.
        """
        vnf_instance_name = "cnf_rollback_after_scale_out_fail"
        vnf_instance_description = "cnf rollback after scale out fail"
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/statefulset_scale.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "simple", vnf_instance_name,
            vnf_instance_description, inst_additional_param)
        # fail scale out for rollback
        aspect_id = "vdu1_aspect"
        previous_level = self._test_cnf_scale(vnf_instance, aspect_id,
                                              number_of_steps=2, error=True)
        # test rollback
        self._test_rollback_cnf_scale(
            vnf_instance['id'], aspect_id, previous_level)
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils


def sub_create_min(callback_uri):
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "callbackUri": callback_uri
    }


def sub_create_max(callback_uri):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    vnf_provider_1 = {
        "vnfProvider": "dummy-vnfProvider-1",
        "vnfProducts": [
            {
                "vnfProductName": "dummy-vnfProductName-1-1",
                "versions": [
                    {
                        "vnfSoftwareVersion": 1.0,
                        "vnfdVersions": [1.0, 2.0]
                    },
                    {
                        "vnfSoftwareVersion": 1.1,
                        "vnfdVersions": [1.1, 2.1]
                    },
                ]
            },
            {
                "vnfProductName": "dummy-vnfProductName-1-2",
                "versions": [
                    {
                        "vnfSoftwareVersion": 1.0,
                        "vnfdVersions": [1.0, 2.0]
                    },
                    {
                        "vnfSoftwareVersion": 1.1,
                        "vnfdVersions": [1.1, 2.1]
                    },
                ]
            }
        ]
    }
    vnf_provider_2 = {
        "vnfProvider": "dummy-vnfProvider-2",
        "vnfProducts": [
            {
                "vnfProductName": "dummy-vnfProductName-2-1",
                "versions": [
                    {
                        "vnfSoftwareVersion": 1.0,
                        "vnfdVersions": [1.0, 2.0]
                    },
                    {
                        "vnfSoftwareVersion": 1.1,
                        "vnfdVersions": [1.1, 2.1]
                    },
                ]
            },
            {
                "vnfProductName": "dummy-vnfProductName-2-2",
                "versions": [
                    {
                        "vnfSoftwareVersion": 1.0,
                        "vnfdVersions": [1.0, 2.0]
                    },
                    {
                        "vnfSoftwareVersion": 1.1,
                        "vnfdVersions": [1.1, 2.1]
                    },
                ]
            }
        ]
    }

    # NOTE: The following is omitted because authType is BASIC in this case
    #  - "paramsOauth2ClientCredentials"
    return {
        "filter": {
            "vnfInstanceSubscriptionFilter": {
                "vnfdIds": [
                    "dummy-vnfdId-1",
                    "dummy-vnfdId-2"
                ],
                "vnfProductsFromProviders": [
                    vnf_provider_1,
                    vnf_provider_2
                ],
                "vnfInstanceIds": [
                    "dummy-vnfInstanceId-1",
                    "dummy-vnfInstanceId-2"
                ],
                "vnfInstanceNames": [
                    "dummy-vnfInstanceName-1",
                    "dummy-vnfInstanceName-2"
                ]
            },
            "notificationTypes": [
                "VnfLcmOperationOccurrenceNotification",
                "VnfIdentifierCreationNotification",
                "VnfLcmOperationOccurrenceNotification"
            ],
            "operationTypes": [
                "INSTANTIATE",
                "SCALE",
                "TERMINATE",
                "HEAL",
                "MODIFY_INFO",
                "CHANGE_EXT_CONN"
            ],
            "operationStates": [
                "COMPLETED",
                "FAILED",
                "FAILED_TEMP",
                "PROCESSING",
                "ROLLING_BACK",
                "ROLLED_BACK",
                "STARTING"
            ]
        },
        "callbackUri": callback_uri,
        "authentication": {
            "authType": [
                "BASIC"
            ],
            "paramsBasic": {
                "password": "test_pass",
                "userName": "test_user"
            },
            # "paramsOauth2ClientCredentials": omitted,
        },
        "verbosity": "SHORT"
    }


def create_vnf_max(vnfd_id):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    return {
        "vnfdId": vnfd_id,
        "vnfInstanceName": "sample",
        "vnfInstanceDescription": "test sample",
        "metadata": {"dummy-key": "dummy-val"}
    }


def terminate_vnf_max():
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    return {
        "terminationType": "GRACEFUL",
        "gracefulTerminationTimeout": 5,
        "additionalParams": {"dummy-key": "dummy-val"}
    }


def instantiate_vnf_max(net_ids, subnets, ports, auth_url):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)

    vim_id_1 = uuidutils.generate_uuid()
    vim_id_2 = uuidutils.generate_uuid()
    link_port_id_1 = uuidutils.generate_uuid()
    link_port_id_2 = uuidutils.generate_uuid()

    # NOTE: The following is not supported so it is omitted
    #  - "segmentationId"
    #  - "addressRange"
    #  - Multiple "cpProtocolData"
    #  - Multiple "fixedAddresses"
    ext_vl_1 = {
        "id": uuidutils.generate_uuid(),
        "vimConnectionId": vim_id_1,
        "resourceProviderId": "Company",
        "resourceId": net_ids['net0'],
        "extCps": [
            {
                "cpdId": "VDU1_CP1",
                "cpConfig": {
                    "VDU1_CP1": {
                        "parentCpConfigId": uuidutils.generate_uuid(),
                        # "linkPortId": omitted,
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                # "macAddress": omitted,
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    # "fixedAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['subnet0']}]}}]
                    },
                    # { "VDU1_CP1_2": omitted }
                }
            },
            {
                "cpdId": "VDU2_CP1-1",
                "cpConfig": {
                    "VDU2_CP1-1": {
                        "parentCpConfigId": uuidutils.generate_uuid(),
                        "linkPortId": link_port_id_1,
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                # "macAddress": omitted,
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    # "fixedAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['subnet0']
                                }]
                            }
                        }]
                    },
                    # { "VDU2_CP1_2": omitted }
                }
            },
            {
                "cpdId": "VDU2_CP1-2",
                "cpConfig": {
                    "VDU2_CP1-2": {
                        "parentCpConfigId": uuidutils.generate_uuid(),
                        "linkPortId": link_port_id_2,
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                # "macAddress": omitted,
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    # "fixedAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['subnet0']
                                }]
                            }
                        }]
                    },
                    # { "VDU2_CP1_2": omitted }
                }
            }
        ],
        "extLinkPorts": [
            {
                "id": link_port_id_1,
                "resourceHandle": {
                    "resourceId": ports['VDU2_CP1-1']
                }
            },
            # NOTE: Set dummy value because it is set by "additionalParams"
            {
                "id": link_port_id_2,
                "resourceHandle": {
                    "resourceId": "dummy-id"
                }
            }
        ]
    }

    # NOTE: The following is not supported so it is omitted
    #  - "segmentationId"
    #  - "addressRange"
    #  - Multiple "cpProtocolData"
    #  - Multiple "fixedAddresses"
    ext_vl_2 = {
        "id": uuidutils.generate_uuid(),
        "vimConnectionId": vim_id_1,
        "resourceProviderId": "Company",
        "resourceId": net_ids['ft-net0'],
        "extCps": [
            {
                "cpdId": "VDU1_CP2",
                "cpConfig": {
                    "VDU1_CP2": {
                        "parentCpConfigId": uuidutils.generate_uuid(),
                        # "linkPortId": omitted,
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                # "macAddress": omitted,
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    # "fixedAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv4-subnet0']}
                                ]}
                        }]
                    },
                    # { "VDU1_CP2_2": omitted }
                }
            },
            {
                "cpdId": "VDU2_CP2",
                "cpConfig": {
                    "VDU2_CP2": {
                        "parentCpConfigId": uuidutils.generate_uuid(),
                        # "linkPortId": omitted,
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                "macAddress": "fa:16:3e:fa:22:75",
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    "fixedAddresses": [
                                        "100.100.100.11",
                                        # omitted
                                    ],
                                    # "numDynamicAddresses": omitted,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv4-subnet0']
                                }, {
                                    "type": "IPV6",
                                    # "fixedAddresses": omitted,
                                    # "numDynamicAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv6-subnet0']
                                }]
                            }
                        }]
                    },
                    # { "VDU2_CP2_2": omitted }
                }
            }
        ]
        # "extLinkPorts": omitted
    }
    # NOTE: "vnfLinkPort" is omitted because it is not supported
    ext_mngd_vl_1 = {
        "id": uuidutils.generate_uuid(),
        "vnfVirtualLinkDescId": "internalVL1",
        "vimConnectionId": vim_id_1,
        "resourceProviderId": "Company",
        "resourceId": net_ids['net_mgmt'],
        # "vnfLinkPort": omitted,
        "extManagedMultisiteVirtualLinkId": uuidutils.generate_uuid()
    }
    # NOTE: "vnfLinkPort" is omitted because it is not supported
    ext_mngd_vl_2 = {
        "id": uuidutils.generate_uuid(),
        "vnfVirtualLinkDescId": "internalVL2",
        "vimConnectionId": vim_id_1,
        "resourceProviderId": "Company",
        "resourceId": net_ids['net1'],
        # "vnfLinkPort": omitted,
        "extManagedMultisiteVirtualLinkId": uuidutils.generate_uuid()
    }
    vim_1 = {
        "vimId": vim_id_1,
        "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
        "interfaceInfo": {"endpoint": auth_url},
        "accessInfo": {
            "username": "nfv_user",
            "region": "RegionOne",
            "password": "devstack",
            "project": "nfv",
            "projectDomain": "Default",
            "userDomain": "Default"
        },
        "extra": {"dummy-key": "dummy-val"}
    }
    vim_2 = {
        "vimId": vim_id_2,
        "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
        "interfaceInfo": {"endpoint": auth_url},
        "accessInfo": {
            "username": "dummy_user",
            "region": "RegionOne",
            "password": "dummy_password",
            "project": "dummy_project",
            "projectDomain": "Default",
            "userDomain": "Default"
        },
        "extra": {"dummy-key": "dummy-val"}
    }
    add_params = {
        "lcm-operation-user-data": "./UserData/userdata.py",
        "lcm-operation-user-data-class": "UserData",
        "nfv": {"CP": {"VDU2_CP1-2": {"port": ports['VDU2_CP1-2']}}}
    }

    return {
        "flavourId": "simple",
        "instantiationLevelId": "instantiation_level_1",
        "extVirtualLinks": [
            ext_vl_1,
            ext_vl_2
        ],
        "extManagedVirtualLinks": [
            ext_mngd_vl_1,
            ext_mngd_vl_2
        ],
        "vimConnectionInfo": {
            "vim1": vim_1,
            "vim2": vim_2
        },
        "localizationLanguage": "ja",
        "additionalParams": add_params,
        "extensions": {"dummy-key": "dummy-val"},
        "vnfConfigurableProperties": {"dummy-key": "dummy-val"}
    }


def create_vnf_min(vnfd_id):
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "vnfdId": vnfd_id,
    }


def terminate_vnf_min():
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "terminationType": "FORCEFUL"
    }


def instantiate_vnf_min():
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "flavourId": "simple"
    }


def scaleout_vnf_max():
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    return {
        "type": "SCALE_OUT",
        "aspectId": "VDU1_scale",
        "numberOfSteps": 1,
        "additionalParams": {"dummy-key": "dummy-value"}
    }


def scaleout_vnf_min():
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "type": "SCALE_OUT",
        "aspectId": "VDU1_scale"
    }


def scalein_vnf_max():
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    return {
        "type": "SCALE_IN",
        "aspectId": "VDU1_scale",
        "numberOfSteps": 1,
        "additionalParams": {"dummy-key": "dummy-value"}
    }


def scalein_vnf_min():
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "type": "SCALE_IN",
        "aspectId": "VDU1_scale"
    }


def update_vnf_max(vnfd_id, vnfc_ids):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    return {
        "vnfInstanceName": "new name",
        "vnfInstanceDescription": "new description",
        "vnfdId": vnfd_id,
        "vnfConfigurableProperties": {"dummy-key": "dummy-value"},
        "metadata": {"dummy-key": "dummy-value"},
        "extensions": {"dummy-key": "dummy-value"},
        "vimConnectionInfo": {
            "vim2": {
                "vimId": "ac2d2ece-5e49-4b15-b92d-b681e9c096d8",
                "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
                "interfaceInfo": {
                    "endpoint": "http://127.0.0.1/identity/v3"
                },
                "accessInfo": {
                    "username": "dummy_user",
                    "region": "RegionOne",
                    "password": "dummy_password",
                    "project": "dummy_project",
                    "projectDomain": "Default",
                    "userDomain": "Default"
                },
                "extra": {
                    "dummy-key": "dummy-val"
                }
            }
        },
        "vnfcInfoModifications": [
            {
                "id": id,
                "vnfcConfigurableProperties": {
                    "dummy-key": "dummy-value"
                }} for id in vnfc_ids]
    }


def update_vnf_min():
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "vnfInstanceName": "new name"
    }


def update_vnf_min_with_parameter(vnfd_id):
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "vnfdId": vnfd_id
    }


def heal_vnf_vnfc_max(vnfc_id):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    return {
        "cause": "ManualHealing",
        "vnfcInstanceId": [vnfc_id],
        "additionalParams": {"dummy-key": "dummy-val"}
    }


# The input parameter is_all is bool type, which accepts only True or False.
def heal_vnf_vnfc_max_with_parameter(vnfc_ids, is_all=None):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    if is_all is not None:
        key = "all"
        value = is_all
    else:
        key = "dummy-key"
        value = "dummy-val"

    return {
        "cause": "ManualHealing",
        "vnfcInstanceId": [vnfc_id for vnfc_id in vnfc_ids],
        "additionalParams": {key: value}
    }


# The input parameter is_all is bool type, which accepts only True or False.
def heal_vnf_all_max_with_parameter(is_all=None):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)
    if is_all is not None:
        key = "all"
        value = is_all
    else:
        key = "dummy-key"
        value = "dummy-val"

    return {
        "cause": "ManualHealing",
        "additionalParams": {key: value}
    }


def heal_vnf_vnfc_min(vnfd_id):
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {
        "vnfcInstanceId": [vnfd_id]
    }


def heal_vnf_all_min():
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)
    return {}


def change_ext_conn_max(net_ids, subnets, auth_url):
    # All attributes are set.
    # NOTE: All of the following cardinality attributes are set.
    # In addition, 0..N or 1..N attributes are set to 2 or more.
    #  - 0..1 (1)
    #  - 0..N (2 or more)
    #  - 1
    #  - 1..N (2 or more)

    vim_id_1 = uuidutils.generate_uuid()
    vim_id_2 = uuidutils.generate_uuid()

    ext_vl_1 = {
        "id": uuidutils.generate_uuid(),
        "vimConnectionId": vim_id_1,
        "resourceProviderId": uuidutils.generate_uuid(),
        "resourceId": net_ids['ft-net1'],
        "extCps": [
            {
                "cpdId": "VDU1_CP1",
                "cpConfig": {
                    "VDU1_CP1": {
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                # "macAddress": omitted,
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    # "fixedAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv4-subnet1']}]
                            }
                        }]}
                }
            },
            {
                "cpdId": "VDU2_CP2",
                "cpConfig": {
                    "VDU2_CP2": {
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                # "macAddress": omitted,
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    "fixedAddresses": [
                                        "22.22.22.101"
                                    ],
                                    # "numDynamicAddresses": omitted
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv4-subnet1']
                                }, {
                                    "type": "IPV6",
                                    # "fixedAddresses": omitted,
                                    # "numDynamicAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv6-subnet1']
                                }]
                            }
                        }]
                    }}
            }
        ]
    }
    vim_1 = {
        "vimId": vim_id_1,
        "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
        "interfaceInfo": {"endpoint": auth_url},
        "accessInfo": {
            "username": "nfv_user",
            "region": "RegionOne",
            "password": "devstack",
            "project": "nfv",
            "projectDomain": "Default",
            "userDomain": "Default"
        },
        "extra": {"dummy-key": "dummy-val"}
    }
    vim_2 = {
        "vimId": vim_id_2,
        "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
        "interfaceInfo": {"endpoint": auth_url},
        "accessInfo": {
            "username": "dummy_user",
            "region": "RegionOne",
            "password": "dummy_password",
            "project": "dummy_project",
            "projectDomain": "Default",
            "userDomain": "Default"
        },
        "extra": {"dummy-key": "dummy-val"}
    }

    return {
        "extVirtualLinks": [
            ext_vl_1
        ],
        "vimConnectionInfo": {
            "vim1": vim_1,
            "vim2": vim_2
        },
        "additionalParams": {"dummy-key": "dummy-val"}
    }


def change_ext_conn_min(net_ids, subnets):
    # Omit except for required attributes
    # NOTE: Only the following cardinality attributes are set.
    #  - 1
    #  - 1..N (1)

    ext_vl_1 = {
        "id": uuidutils.generate_uuid(),
        "resourceId": net_ids['ft-net1'],
        "extCps": [
            {
                "cpdId": "VDU2_CP2",
                "cpConfig": {
                    "VDU2_CP2": {
                        "cpProtocolData": [{
                            "layerProtocol": "IP_OVER_ETHERNET",
                            "ipOverEthernet": {
                                # "macAddress": omitted,
                                # "segmentationId": omitted,
                                "ipAddresses": [{
                                    "type": "IPV4",
                                    "fixedAddresses": [
                                        "22.22.22.100"
                                    ],
                                    # "numDynamicAddresses": omitted
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv4-subnet1']
                                }, {
                                    "type": "IPV6",
                                    # "fixedAddresses": omitted,
                                    # "numDynamicAddresses": omitted,
                                    "numDynamicAddresses": 1,
                                    # "addressRange": omitted,
                                    "subnetId": subnets['ft-ipv6-subnet1']
                                }]
                            }
                        }]}
                }
            }
        ]
    }

    return {
        "extVirtualLinks": [
            ext_vl_1
        ]
    }



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

from keystoneauth1 import exceptions
from oslo_config import cfg

from tacker import context as t_context
from tacker.extensions import nfvo
from tacker.nfvo.drivers.vim import openstack_driver
from tacker.tests.unit import base
from tacker.tests.unit.db import utils

OPTS = [cfg.StrOpt('user_domain_id',
                   default='default',
                   help='User Domain Id'),
        cfg.StrOpt('project_domain_id',
                   default='default',
                   help='Project Domain Id'),
        cfg.StrOpt('password',
                   default='default',
                   help='User Password'),
        cfg.StrOpt('username',
                   default='default',
                   help='User Name'),
        cfg.StrOpt('user_domain_name',
                   default='default',
                   help='Use Domain Name'),
        cfg.StrOpt('project_name',
                   default='default',
                   help='Project Name'),
        cfg.StrOpt('project_domain_name',
                   default='default',
                   help='Project Domain Name'),
        cfg.StrOpt('auth_url',
                   default='http://localhost/identity/v3',
                   help='Keystone endpoint')]

cfg.CONF.register_opts(OPTS, 'keystone_authtoken')
CONF = cfg.CONF


class FakeKeystone(mock.Mock):
    pass


class FakeNeutronClient(mock.Mock):
    pass


class FakeKeymgrAPI(mock.Mock):
    pass


class mock_dict(dict):
    def __getattr__(self, item):
        return self.get(item)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class TestOpenstack_Driver(base.TestCase):
    def setUp(self):
        super(TestOpenstack_Driver, self).setUp()
        self._mock_keystone()
        self.keystone.create_key_dir.return_value = 'test_keys'
        self.config_fixture.config(group='vim_keys', openstack='/tmp/')
        self.config_fixture.config(group='vim_keys', use_barbican=False)
        self.openstack_driver = openstack_driver.OpenStack_Driver()
        self.vim_obj = self.get_vim_obj()
        self.auth_obj = utils.get_vim_auth_obj()
        self.addCleanup(mock.patch.stopall)
        self._mock_keymgr()

    def _mock_keystone(self):
        self.keystone = mock.Mock(wraps=FakeKeystone())
        fake_keystone = mock.Mock()
        fake_keystone.return_value = self.keystone
        self._mock(
            'tacker.vnfm.keystone.Keystone', fake_keystone)

    def _mock_keymgr(self):
        self.keymgr = mock.Mock(wraps=FakeKeymgrAPI())
        fake_keymgr = mock.Mock()
        fake_keymgr.return_value = self.keymgr
        self._mock(
            'tacker.keymgr.barbican_key_manager.BarbicanKeyManager',
            fake_keymgr)

    def get_vim_obj(self):
        return {'id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff', 'type':
                'openstack', 'auth_url': 'http://localhost/identity',
                'auth_cred': {'username': 'test_user',
                              'password': 'test_password',
                              'user_domain_name': 'default',
                              'cert_verify': 'True',
                              'auth_url': 'http://localhost/identity'},
                'name': 'VIM0',
                'vim_project': {'name': 'test_project',
                                'project_domain_name': 'default'}}

    def get_vim_obj_barbican(self):
        return {'id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff', 'type':
                'openstack', 'auth_url': 'http://localhost/identity',
                'auth_cred': {'username': 'test_user',
                              'password': 'test_password',
                              'user_domain_name': 'default',
                              'key_type': 'barbican_key',
                              'secret_uuid': 'fake-secret-uuid',
                              'cert_verify': 'True',
                              'auth_url': 'http://localhost/identity'},
                'name': 'VIM0',
                'vim_project': {'name': 'test_project',
                                'project_domain_name': 'default'}}

    def test_register_keystone_v3(self):
        regions = [mock_dict({'id': 'RegionOne'})]
        attrs = {'regions.list.return_value': regions}
        keystone_version = 'v3'
        mock_ks_client = mock.Mock(**attrs)
        self.keystone.get_version.return_value = keystone_version
        self._test_register_vim(self.vim_obj, mock_ks_client)
        mock_ks_client.regions.list.assert_called_once_with()
        self.keystone.initialize_client.assert_called_once_with(
            **self.auth_obj)

    def _test_register_vim(self, vim_obj, mock_ks_client):
        self.keystone.initialize_client.return_value = mock_ks_client
        fernet_attrs = {'encrypt.return_value': 'encrypted_password'}
        mock_fernet_obj = mock.Mock(**fernet_attrs)
        mock_fernet_key = b'test_fernet_key'
        self.keystone.create_fernet_key.return_value = (mock_fernet_key,
                                                        mock_fernet_obj)
        self.openstack_driver.register_vim(vim_obj)
        with open('/tmp/' + vim_obj['id'], 'r') as f:
            # asserting that file has been written correctly.
            self.assertEqual('test_fernet_key', f.read())
        mock_fernet_obj.encrypt.assert_called_once_with(mock.ANY)
        os.remove('/tmp/' + vim_obj['id'])

    @mock.patch('tacker.nfvo.drivers.vim.openstack_driver.os.remove')
    @mock.patch('tacker.nfvo.drivers.vim.openstack_driver.os.path'
                '.join')
    def test_deregister_vim(self, mock_os_path, mock_os_remove):
        vim_obj = self.get_vim_obj()
        vim_id = 'my_id'
        vim_obj['id'] = vim_id
        file_path = CONF.vim_keys.openstack + '/' + vim_id
        mock_os_path.return_value = file_path
        self.openstack_driver.deregister_vim(vim_obj)
        mock_os_remove.assert_called_once_with(file_path)

    def test_deregister_vim_barbican(self):
        self.keymgr.delete.return_value = None
        vim_obj = self.get_vim_obj_barbican()
        self.openstack_driver.deregister_vim(vim_obj)
        self.keymgr.delete.assert_called_once_with(
            t_context.generate_tacker_service_context(), 'fake-secret-uuid')

    def test_encode_vim_auth_barbican(self):
        self.config_fixture.config(group='vim_keys',
                                   use_barbican=True)
        fernet_attrs = {'encrypt.return_value': 'encrypted_password'}
        mock_fernet_obj = mock.Mock(**fernet_attrs)
        mock_fernet_key = 'test_fernet_key'
        self.keymgr.store.return_value = 'fake-secret-uuid'
        self.keystone.create_fernet_key.return_value = (mock_fernet_key,
                                                        mock_fernet_obj)

        vim_obj = self.get_vim_obj()
        self.openstack_driver.encode_vim_auth(
            vim_obj['id'], vim_obj['auth_cred'])

        self.keymgr.store.assert_called_once_with(
            t_context.generate_tacker_service_context(), 'test_fernet_key')
        mock_fernet_obj.encrypt.assert_called_once_with(mock.ANY)
        self.assertEqual(vim_obj['auth_cred']['key_type'],
                         'barbican_key')
        self.assertEqual(vim_obj['auth_cred']['secret_uuid'],
                         'fake-secret-uuid')

    def test_register_vim_invalid_auth(self):
        attrs = {'regions.list.side_effect': exceptions.Unauthorized}
        self._test_register_vim_auth(attrs)

    def test_register_vim_missing_auth(self):
        attrs = {'regions.list.side_effect': exceptions.BadRequest}
        self._test_register_vim_auth(attrs)

    def _test_register_vim_auth(self, attrs):
        keystone_version = 'v3'
        self.keystone.get_version.return_value = keystone_version
        mock_ks_client = mock.Mock(**attrs)
        self.keystone.initialize_client.return_value = mock_ks_client
        self.assertRaises(nfvo.VimUnauthorizedException,
                          self.openstack_driver.register_vim,
                          self.vim_obj)
        mock_ks_client.regions.list.assert_called_once_with()
        self.keystone.initialize_client.assert_called_once_with(
            **self.auth_obj)

    def test_get_vim_resource_id(self):
        resource_type = 'network'
        resource_name = 'net0'
        fake_networks = {'networks': [{'id': 'fake-uuid', 'name': 'net0'}]}
        fake_neutron_client = FakeNeutronClient()
        fake_neutron_client.list_networks.return_value = fake_networks
        self.openstack_driver._get_client = mock.Mock(
            return_value=fake_neutron_client)

        self.openstack_driver.get_vim_resource_id(
            self.vim_obj, resource_type, resource_name)

        self.openstack_driver._get_client.assert_called_once_with(
            self.vim_obj, mock.ANY)
        fake_neutron_client.list_networks.assert_called_once_with(
            **{'name': 'net0'})

    def test_get_vim_resource_id_name_not_unique(self):
        resource_type = 'network'
        resource_name = 'net0'
        fake_networks = {'networks': [{'id': 'fake-uuid-1', 'name': 'net0'},
                                      {'id': 'fake-uuid-2', 'name': 'net0'}]}
        fake_neutron_client = FakeNeutronClient()
        fake_neutron_client.list_networks.return_value = fake_networks
        self.openstack_driver._get_client = mock.Mock(
            return_value=fake_neutron_client)

        self.assertRaises(nfvo.VimGetResourceNameNotUnique,
                          self.openstack_driver.get_vim_resource_id,
                          self.vim_obj, resource_type, resource_name)

    def test_get_vim_resource_id_name_not_exist(self):
        resource_type = 'network'
        resource_name = 'net0'
        fake_networks = {'networks': []}
        fake_neutron_client = FakeNeutronClient()
        fake_neutron_client.list_networks.return_value = fake_networks
        self.openstack_driver._get_client = mock.Mock(
            return_value=fake_neutron_client)

        self.assertRaises(nfvo.VimGetResourceNotFoundException,
                          self.openstack_driver.get_vim_resource_id,
                          self.vim_obj, resource_type, resource_name)



# Copyright (c) 2013 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from tacker.common import log as call_log
from tacker.tests import base


MODULE_NAME = 'tacker.tests.unit.test_common_log'


class TargetKlass(object):

    @call_log.log
    def test_method(self, arg1, arg2, *args, **kwargs):
        pass


class TestCallLog(base.BaseTestCase):
    def setUp(self):
        super(TestCallLog, self).setUp()
        self.skipTest("Not ready yet")
        self.klass = TargetKlass()
        self.expected_format = ('%(class_name)s method %(method_name)s '
                                'called with arguments %(args)s %(kwargs)s')
        self.expected_data = {'class_name': MODULE_NAME + '.TargetKlass',
                              'method_name': 'test_method',
                              'args': (),
                              'kwargs': {}}

    def test_call_log_all_args(self):
        self.expected_data['args'] = (10, 20)
        with mock.patch.object(call_log.LOG, 'debug') as log_debug:
            self.klass.test_method(10, 20)
            log_debug.assert_called_once_with(self.expected_format,
                                              self.expected_data)

    def test_call_log_all_kwargs(self):
        self.expected_data['kwargs'] = {'arg1': 10, 'arg2': 20}
        with mock.patch.object(call_log.LOG, 'debug') as log_debug:
            self.klass.test_method(arg1=10, arg2=20)
            log_debug.assert_called_once_with(self.expected_format,
                                              self.expected_data)

    def test_call_log_known_args_unknown_args_kwargs(self):
        self.expected_data['args'] = (10, 20, 30)
        self.expected_data['kwargs'] = {'arg4': 40}
        with mock.patch.object(call_log.LOG, 'debug') as log_debug:
            self.klass.test_method(10, 20, 30, arg4=40)
            log_debug.assert_called_once_with(self.expected_format,
                                              self.expected_data)

    def test_call_log_known_args_kwargs_unknown_kwargs(self):
        self.expected_data['args'] = (10,)
        self.expected_data['kwargs'] = {'arg2': 20, 'arg3': 30, 'arg4': 40}
        with mock.patch.object(call_log.LOG, 'debug') as log_debug:
            self.klass.test_method(10, arg2=20, arg3=30, arg4=40)
            log_debug.assert_called_once_with(self.expected_format,
                                              self.expected_data)

    def test_call_log_password_mask_args_kwargs(self):
        auth_cred = {'userame': 'demo', 'password': 'changeit'}
        self.expected_data['kwargs'] = {'password': '***'}
        self.expected_data['args'] = ({'userame': 'demo', 'password': '***'})
        with mock.patch.object(call_log.LOG, 'debug') as log_debug:
            self.klass.test_method(auth_cred, password='guessme')
            log_debug.assert_called_once_with(self.expected_format,
                                              self.expected_data)



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from sqlalchemy import desc
from sqlalchemy.orm import joinedload

from tacker.common import exceptions
from tacker import context as t_context
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.objects import fields
from tacker.tests.functional import base
from tacker.tests import utils

VNF_PACKAGE_UPLOAD_TIMEOUT = 300
VNF_INSTANTIATE_TIMEOUT = 600
VNF_TERMINATE_TIMEOUT = 600
RETRY_WAIT_TIME = 5


def _create_and_upload_vnf_package(tacker_client, csar_package_name,
                                   user_defined_data):
    # create vnf package
    body = jsonutils.dumps({"userDefinedData": user_defined_data})
    resp, vnf_package = tacker_client.do_request(
        '/vnfpkgm/v1/vnf_packages', "POST", body=body)

    # upload vnf package
    csar_package_path = "../../../etc/samples/etsi/nfv/%s" % csar_package_name
    file_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                             csar_package_path))

    # Generating unique vnfd id. This is required when multiple workers
    # are running concurrently. The call below creates a new temporary
    # CSAR with unique vnfd id.
    file_path, uniqueid = utils.create_csar_with_unique_vnfd_id(file_path)

    with open(file_path, 'rb') as file_object:
        resp, resp_body = tacker_client.do_request(
            '/vnfpkgm/v1/vnf_packages/{id}/package_content'.format(
                id=vnf_package['id']),
            "PUT", body=file_object, content_type='application/zip')

    # wait for onboard
    timeout = VNF_PACKAGE_UPLOAD_TIMEOUT
    start_time = int(time.time())
    show_url = os.path.join('/vnfpkgm/v1/vnf_packages', vnf_package['id'])
    vnfd_id = None
    while True:
        resp, body = tacker_client.do_request(show_url, "GET")
        if body['onboardingState'] == "ONBOARDED":
            vnfd_id = body['vnfdId']
            break

        if ((int(time.time()) - start_time) > timeout):
            raise Exception("Failed to onboard vnf package")

        time.sleep(1)

    # remove temporarily created CSAR file
    os.remove(file_path)
    return vnf_package['id'], vnfd_id


class VnfLcmTest(base.BaseTackerTest):

    @classmethod
    def setUpClass(cls):
        cls.tacker_client = base.BaseTackerTest.tacker_http_client()

        cls.vnf_package, cls.vnfd_id = \
            _create_and_upload_vnf_package(
                cls.tacker_client, "test_inst_terminate_vnf_with_vnflcmnoop",
                {"key": "file_functional"})

        super(VnfLcmTest, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        # Update vnf package operational state to DISABLED
        update_req_body = jsonutils.dumps({
            "operationalState": "DISABLED"})
        base_path = "/vnfpkgm/v1/vnf_packages"
        for package_id in [cls.vnf_package]:
            resp, resp_body = cls.tacker_client.do_request(
                '{base_path}/{id}'.format(id=package_id,
                                          base_path=base_path),
                "PATCH", content_type='application/json',
                body=update_req_body)

            # Delete vnf package
            url = '/vnfpkgm/v1/vnf_packages/%s' % package_id
            cls.tacker_client.do_request(url, "DELETE")

        super(VnfLcmTest, cls).tearDownClass()

    def setUp(self):
        super(VnfLcmTest, self).setUp()
        self.base_url = "/vnflcm/v1/vnf_instances"
        self.context = t_context.get_admin_context()

        vim_list = self.client.list_vims()
        if not vim_list:
            self.skipTest("Vims are not configured")

        vim_id = 'VIM0'
        vim = self.get_vim(vim_list, vim_id)
        if not vim:
            self.skipTest("Openstack VIM '%s' is missing" % vim_id)
        self.vim_id = vim['id']

    def _instantiate_vnf_instance_request(
            self, flavour_id, vim_id=None, additional_param=None):
        request_body = {"flavourId": flavour_id}

        if vim_id:
            request_body["vimConnectionInfo"] = [
                {"id": uuidutils.generate_uuid(),
                 "vimId": vim_id,
                 "vimType": "openstack"}]

        if additional_param:
            request_body["additionalParams"] = additional_param

        return request_body

    def _create_vnf_instance(self, vnfd_id, vnf_instance_name=None,
                             vnf_instance_description=None):
        request_body = {'vnfdId': vnfd_id}
        if vnf_instance_name:
            request_body['vnfInstanceName'] = vnf_instance_name

        if vnf_instance_description:
            request_body['vnfInstanceDescription'] = vnf_instance_description

        resp, response_body = self.http_client.do_request(
            self.base_url, "POST", body=jsonutils.dumps(request_body))
        return resp, response_body

    def _delete_wait_vnf_instance(self, id):
        timeout = VNF_TERMINATE_TIMEOUT
        url = os.path.join(self.base_url, id)
        start_time = int(time.time())
        while True:
            resp, body = self.http_client.do_request(url, "DELETE")
            if 204 == resp.status_code:
                break

            if ((int(time.time()) - start_time) > timeout):
                error = "Failed to delete vnf instance %s"
                self.fail(error % id)

            time.sleep(RETRY_WAIT_TIME)

    def _delete_vnf_instance(self, id):
        self._delete_wait_vnf_instance(id)

        # verify vnf instance is deleted
        url = os.path.join(self.base_url, id)
        resp, body = self.http_client.do_request(url, "GET")
        self.assertEqual(404, resp.status_code)

    def _show_vnf_instance(self, id, expected_result=None):
        show_url = os.path.join(self.base_url, id)
        resp, vnf_instance = self.http_client.do_request(show_url, "GET")
        self.assertEqual(200, resp.status_code)

        if expected_result:
            self.assertDictSupersetOf(expected_result, vnf_instance)

        return vnf_instance

    def _vnf_instance_wait(
            self, id,
            instantiation_state=fields.VnfInstanceState.INSTANTIATED,
            timeout=VNF_INSTANTIATE_TIMEOUT):
        show_url = os.path.join(self.base_url, id)
        start_time = int(time.time())
        while True:
            resp, body = self.http_client.do_request(show_url, "GET")
            if body['instantiationState'] == instantiation_state:
                break

            if ((int(time.time()) - start_time) > timeout):
                error = ("Vnf instance %(id)s status is %(current)s, "
                         "expected status should be %(expected)s")
                self.fail(error % {"id": id,
                                   "current": body['instantiationState'],
                                   "expected": instantiation_state})

            time.sleep(RETRY_WAIT_TIME)

    def _instantiate_vnf_instance(self, id, request_body):
        url = os.path.join(self.base_url, id, "instantiate")
        resp, body = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)
        self._vnf_instance_wait(id)

    def _terminate_vnf_instance(self, id, request_body):
        url = os.path.join(self.base_url, id, "terminate")
        resp, body = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

        timeout = request_body.get('gracefulTerminationTimeout')
        start_time = int(time.time())

        self._vnf_instance_wait(
            id, instantiation_state=fields.VnfInstanceState.NOT_INSTANTIATED,
            timeout=VNF_TERMINATE_TIMEOUT)

        # If gracefulTerminationTimeout is set, check whether vnf
        # instantiation_state is set to NOT_INSTANTIATED after
        # gracefulTerminationTimeout seconds.
        if timeout and int(time.time()) - start_time < timeout:
            self.fail("Vnf is terminated before graceful termination "
                      "timeout period")

    @db_api.context_manager.reader
    def _vnf_notify_get_by_id(self, context, vnf_instance_id,
                              columns_to_join=None):
        query = api.model_query(
            context, models.VnfLcmOpOccs,
            read_deleted="no", project_only=True).filter_by(
            vnf_instance_id=vnf_instance_id).order_by(
            desc("created_at"))

        if columns_to_join:
            for column in columns_to_join:
                query = query.options(joinedload(column))

        result = query.first()

        if not result:
            raise exceptions.VnfInstanceNotFound(id=vnf_instance_id)

        return result

    def _wait_vnflcm_op_occs(
            self, context, vnf_instance_id, type, timeout,
            operation_state='COMPLETED'):
        start_time = int(time.time())
        while True:
            vnflcm_op_occ = self._vnf_notify_get_by_id(
                context, vnf_instance_id)

            if vnflcm_op_occ.operation_state == operation_state:
                break

            if ((int(time.time()) - start_time) > timeout):
                raise Exception("Failed to wait {} instance".format(type))

            time.sleep(RETRY_WAIT_TIME)

    def test_instantiate_terminate_vnf_with_vnflcmnoop(self):
        # create vnf instance
        vnf_instance_name = "vnf_with_instantiation_level-%s" % \
                            uuidutils.generate_uuid()
        vnf_instance_description = "vnf with instantiation level 1"
        resp, vnf_instance = self._create_vnf_instance(
            self.vnfd_id,
            vnf_instance_name=vnf_instance_name,
            vnf_instance_description=vnf_instance_description)
        self.assertIsNotNone(vnf_instance['id'])
        self.assertEqual(201, resp.status_code)
        # instantiate vnf instance
        request_body = self._instantiate_vnf_instance_request(
            "simple", vim_id=self.vim_id)
        self._instantiate_vnf_instance(vnf_instance['id'], request_body)
        self._wait_vnflcm_op_occs(self.context, vnf_instance['id'],
                                  'instantiate', VNF_INSTANTIATE_TIMEOUT)
        # show vnf instance
        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        self.assertEqual(vnf_instance['instantiationState'], 'INSTANTIATED')
        vnflcm_op_occ_ins = self._vnf_notify_get_by_id(
            self.context, vnf_instance['id'], columns_to_join=None)
        self.assertEqual(vnflcm_op_occ_ins.operation, 'INSTANTIATE')
        # terminate vnf instance
        terminate_req_body = {
            "terminationType": fields.VnfInstanceTerminationType.FORCEFUL,
        }
        self._terminate_vnf_instance(vnf_instance['id'], terminate_req_body)
        self._wait_vnflcm_op_occs(self.context, vnf_instance['id'],
                                  'terminate', VNF_TERMINATE_TIMEOUT)
        vnflcm_op_occ_term = self._vnf_notify_get_by_id(
            self.context, vnf_instance['id'], columns_to_join=None)
        self.assertEqual(vnflcm_op_occ_term.operation, 'TERMINATE')
        # delete vnf instance
        self._delete_vnf_instance(vnf_instance['id'])



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import iso8601

from tacker.db.db_sqlalchemy import models
from tacker.tests import constants
from tacker.tests import uuidsentinel

vnf_package_data = {'algorithm': None, 'hash': None,
                    'location_glance_store': None,
                    'onboarding_state': 'CREATED',
                    'operational_state': 'DISABLED',
                    'tenant_id': uuidsentinel.tenant_id,
                    'usage_state': 'NOT_IN_USE',
                    'user_data': {'abc': 'xyz'},
                    'created_at': datetime.datetime(
                        2019, 8, 8, 0, 0, 0, tzinfo=iso8601.UTC),
                    'deleted': False,
                    'size': 0
                    }

software_image = {
    'software_image_id': uuidsentinel.software_image_id,
    'name': 'test', 'provider': 'test', 'version': 'test',
    'algorithm': 'sha-256',
    'hash': 'b9c3036539fd7a5f87a1bf38eb05fdde8b556a1'
            'a7e664dbeda90ed3cd74b4f9d',
    'container_format': 'test', 'disk_format': 'qcow2', 'min_disk': 1,
    'min_ram': 2, 'size': 1, 'image_path': 'test',
    'metadata': {'key1': 'value1'}
}

artifact_data = {
    'file_name': 'test', 'type': 'test',
    'algorithm': 'sha-256',
    'hash': 'b9c3036539fd7a5f87a1bf38eb05fdde8b556a1'
            'a7e664dbeda90ed3cd74b4f9d',
    'metadata': {'key1': 'value1'}
}

artifacts = {
    'json_data': 'test data',
    'type': 'tosca.artifacts.nfv.SwImage',
    'algorithm': 'sha512', 'hash': uuidsentinel.hash}

filter = {
    "usageState": ["NOT_IN_USE"],
    "vnfPkgId": ["f04857cb-abdc-405f-8254-01501f3fa059"],
    "vnfdId": ["b1bb0ce7-5555-0001-95ed-4840d70a1209"],
    "vnfInstanceSubscriptionFilter": {"vnfdIds": []},
    "vnfProductsFromProviders": [
        {
            "vnfProvider": "xxxxx",
            "vnfProducts": [
                {
                    "vnfProductName": "artifactVNF",
                    "versions": [
                        {
                            "vnfSoftwareVersion": "1.0",
                            "vnfdVersions": ["v2.2"]
                        }
                    ]
                }
            ]
        },
        {
            "vnfProvider": "xxxxx",
            "vnfProducts": [
                {
                    "vnfProductName": "artifactVNF",
                    "versions": [
                        {
                            "vnfSoftwareVersion": "1.0",
                            "vnfdVersions": ["v2.2"]
                        }
                    ]
                }
            ]
        }
    ],
    "notificationTypes": ["VnfLcmOperationOccurrenceNotification"],
    "operationalState": ["ENABLED"],
    "tenant_id": uuidsentinel.tenant_id
}

subscription_data = {
    'id': "c3e5ea85-8e3d-42df-a636-3b7857cbd7f9",
    'callback_uri': "fake_url",
    'created_at': "2020-06-11 09:39:58",
    'tenant_id': uuidsentinel.tenant_id
}

vnfd_data = {
    "tenant_id": uuidsentinel.tenant_id,
    'name': 'test',
    'description': 'test_description',
    'mgmt_driver': 'test_mgmt_driver'
}

vnfd_attribute = {
    'key': 'test_key',
    'value': 'test_value',
}

lcm_op_occs_data = {
    "tenant_id": uuidsentinel.tenant_id,
    'operation_state': 'PROCESSING',
    'state_entered_time': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                            tzinfo=iso8601.UTC),
    'start_time': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                    tzinfo=iso8601.UTC),
    'operation': 'MODIFY_INFO',
    'is_automatic_invocation': 0,
    'is_cancel_pending': 0,
}

vim_data = {
    'id': uuidsentinel.vim_id,
    'type': 'test_type',
    "tenant_id": uuidsentinel.tenant_id,
    'name': "test_name",
    'description': "test_description",
    'placement_attr': "test_placement_attr",
    'shared': 0,
    'status': "REACHABLE",
    'is_default': 0
}

fake_vnf_package_response = copy.deepcopy(vnf_package_data)
fake_vnf_package_response.pop('user_data')
fake_vnf_package_response.update({'id': uuidsentinel.package_uuid})

vnf_deployment_flavour = {'flavour_id': 'simple',
                          'flavour_description': 'simple flavour description',
                          'instantiation_levels': {
                              'levels': {
                                  'instantiation_level_1': {
                                      'description': 'Smallest size',
                                      'scale_info': {
                                          'worker_instance': {
                                              'scale_level': 0
                                          }
                                      }
                                  },
                                  'instantiation_level_2': {
                                      'description': 'Largest size',
                                      'scale_info': {
                                          'worker_instance': {
                                              'scale_level': 2
                                          }
                                      }
                                  }
                              },
                              'default_level': 'instantiation_level_1'
                          },
                          'created_at': datetime.datetime(
                              2019, 8, 8, 0, 0, 0, tzinfo=iso8601.UTC),
                          }

vnf_artifacts = {
    'artifact_path': 'scripts/install.sh',
    '_metadata': {},
    'algorithm': 'sha-256',
    'hash': 'd0e7828293355a07c2dccaaa765c80b507e60e6167067c950dc2e6b0da0dbd8b',
    'created_at': datetime.datetime(2020, 6, 29, 0, 0, 0, tzinfo=iso8601.UTC),
}


def fake_vnf_package_vnfd_dict(**updates):
    vnf_pkg_vnfd = {
        'package_uuid': uuidsentinel.package_uuid,
        'vnfd_id': uuidsentinel.vnfd_id,
        'vnf_provider': 'test vnf provider',
        'vnf_product_name': 'Sample VNF',
        'vnf_software_version': '1.0',
        'vnfd_version': '1.0'
    }

    if updates:
        vnf_pkg_vnfd.update(updates)

    return vnf_pkg_vnfd


def return_vnf_package_vnfd_data():
    model_obj = models.VnfPackageVnfd()
    model_obj.update(fake_vnf_package_vnfd_dict())
    return model_obj


def get_vnf_package_vnfd_data(vnf_package_id, vnfd_id):
    return {
        'package_uuid': vnf_package_id,
        'vnfd_id': vnfd_id,
        'vnf_provider': 'test vnf provider',
        'vnf_product_name': 'Sample VNF',
        'vnf_software_version': '1.0',
        "vnf_pkg_id": uuidsentinel.vnf_pkg_id,
        'vnfd_version': '1.0',
    }


def get_vnf_instance_data(vnfd_id):
    return {
        "vnf_software_version": "1.0",
        "vnf_product_name": "Sample VNF",
        "vnf_instance_name": 'Sample VNF Instance',
        "vnf_instance_description": 'Sample vnf_instance_description',
        "instantiation_state": "NOT_INSTANTIATED",
        "vnf_provider": "test vnf provider",
        "vnfd_id": vnfd_id,
        "vnfd_version": "1.0",
        "tenant_id": uuidsentinel.tenant_id,
        "vnf_pkg_id": uuidsentinel.vnf_pkg_id,
        "vnf_metadata": {"key": "value"},
    }


def get_vnf_instance_data_with_id(vnfd_id):
    return {
        "id": uuidsentinel.tenant_id,
        "vnf_software_version": "1.0",
        "vnf_product_name": "Sample VNF",
        "vnf_instance_name": 'Sample VNF Instance',
        "vnf_instance_description": 'Sample vnf_instance_description',
        "instantiation_state": "NOT_INSTANTIATED",
        "vnf_provider": "test vnf provider",
        "vnfd_id": vnfd_id,
        "vnfd_version": "1.0",
        "tenant_id": uuidsentinel.tenant_id,
        "vnf_pkg_id": uuidsentinel.vnf_pkg_id,
        "vnf_metadata": {"key": "value"},
    }


def get_lcm_op_occs_data(id, vnf_instance_id):
    return {
        "id": id,
        "tenant_id": uuidsentinel.tenant_id,
        'operation_state': 'PROCESSING',
        'state_entered_time': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                                tzinfo=iso8601.UTC),
        'start_time': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'vnf_instance_id': vnf_instance_id,
        'operation': 'MODIFY_INFO',
        'is_automatic_invocation': 0,
        'is_cancel_pending': 0,
    }


def fake_vnf_instance_model_dict(**updates):
    vnf_instance = {
        'deleted': False,
        'deleted_at': None,
        'updated_at': None,
        'created_at': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'vnf_product_name': 'Sample VNF',
        'vnf_instance_name': 'Sample VNF',
        'vnf_instance_description': None,
        'vnf_provider': 'test vnf provider',
        'vnf_software_version': '1.0',
        'vnfd_id': uuidsentinel.vnfd_id,
        'vnfd_version': '1.0',
        'instantiation_state': 'NOT_INSTANTIATED',
        'vim_connection_info': [],
        'tenant_id': '33f8dbdae36142eebf214c1869eb4e4c',
        'vnf_pkg_id': uuidsentinel.vnf_pkg_id,
        'id': constants.UUID,
        'vnf_metadata': {"key": "value"},
    }

    if updates:
        vnf_instance.update(updates)

    return vnf_instance


def fake_vnf_resource_data(instance_id):

    return {
        'vnf_instance_id': instance_id,
        'resource_name': "test",
        'resource_type': "image",
        'resource_identifier': uuidsentinel.image_id,
        'resource_status': "status",
        'tenant_id': uuidsentinel.tenant_id
    }


def vnf_pack_vnfd_data(vnf_pack_id):
    return {
        'package_uuid': vnf_pack_id,
        'vnfd_id': uuidsentinel.vnfd_id,
        'vnf_provider': 'test_provider',
        'vnf_product_name': 'test_product_name',
        'vnf_software_version': 'test_version',
        'vnfd_version': 'test_vnfd_version',
    }


def vnf_pack_artifact_data(vnf_pack_id):
    return {
        'package_uuid': vnf_pack_id,
        'artifact_path': 'scripts/install.sh',
        'algorithm': 'sha-256',
        'hash': 'd0e7828293355a07c2dccaaa765c80b507e'
                '60e6167067c950dc2e6b0da0dbd8b',
        '_metadata': {}
    }


ip_address = [{
    'type': 'IPV4',
    'is_dynamic': True
}]

ip_over_ethernet_address_info = {
    'mac_address': 'fake_mac',
    'ip_addresses': ip_address,
}

cp_protocol_info = {
    'layer_protocol': 'IP_OVER_ETHERNET',
    'ip_over_ethernet': ip_over_ethernet_address_info,
}

vnf_external_cp_info = {
    'id': uuidsentinel.external_cp_id,
    'cpd_id': uuidsentinel.cpd_id,
    'ext_link_port_id': uuidsentinel.ext_link_port_id
}

resource_handle_info = {
    'resource_id': uuidsentinel.resource_id,
    'vim_level_resource_type': 'TEST'
}

ext_link_port_info = {
    'id': uuidsentinel.ext_link_port_id,
    'resource_handle': resource_handle_info,
    'cp_instance_id': uuidsentinel.cp_instance_id,
}

ext_virtual_link_info = {
    'id': uuidsentinel.virtual_link_id,
    'resource_handle': resource_handle_info,
    'ext_link_ports': [ext_link_port_info],
}

vnf_link_ports = {
    'id': uuidsentinel.vnf_link_ports_id,
    'resource_handle': resource_handle_info,
    'cp_instance_id': uuidsentinel.cp_instance_id
}

ext_managed_virtual_link_info = {
    'id': uuidsentinel.ext_managed_virtual_link_id,
    'vnf_virtual_link_desc_id': uuidsentinel.vnf_virtual_link_desc_id,
    'network_resource': resource_handle_info,
    'vnf_link_ports': [vnf_link_ports],
}

vnfc_resource_info = {
    'id': uuidsentinel.resource_info_id,
    'vdu_id': 'vdu1',
    'compute_resource': None,
    'storage_resource_ids': [uuidsentinel.id1, uuidsentinel.id2],
    'reservation_id': uuidsentinel.reservation_id,
    'vnfc_cp_info': None,
    'metadata': {'key': 'value'}

}

vnfc_cp_info = {
    'id': uuidsentinel.cp_instance_id,
    'cpd_id': uuidsentinel.cpd_id,
    'vnf_ext_cp_id': uuidsentinel.vnf_ext_cp_id,
    'cp_protocol_info': [cp_protocol_info],
    'vnf_link_port_id': uuidsentinel.vnf_link_port_id,
}

vnfc_resource_info = {
    'id': uuidsentinel.resource_info_id,
    'vdu_id': uuidsentinel.vdu_id,
    'compute_resource': resource_handle_info,
    'storage_resource_ids': [uuidsentinel.id1, uuidsentinel.id2],
    'reservation_id': uuidsentinel.reservation_id,
    'vnfc_cp_info': [vnfc_cp_info],
    'metadata': {'key': 'value'}
}

ip_address_info = {
    'type': 'IPV4',
    'subnet_id': uuidsentinel.subnet_id,
    'is_dynamic': False,
    'addresses': ['10.10.1', '10.10.2'],
}

vnf_virtual_link_resource_info = {
    'id': uuidsentinel.virtual_link_resource_id,
    'vnf_virtual_link_desc_id': uuidsentinel.vnf_virtual_link_desc_id,
    'network_resource': resource_handle_info,
    'vnf_link_ports': vnf_link_ports,
}

virtual_storage_resource_info = {
    'id': uuidsentinel.virtual_storage_resource_id,
    'virtual_storage_desc_id': uuidsentinel.virtual_storage_desc_id,
    'storage_resource': resource_handle_info,
}

vnf_ext_cp_info = {
    'id': uuidsentinel.id,
    'cpd_id': 'CP1',
    'cp_protocol_info': [cp_protocol_info],
    'associated_vnfc_cp_id': uuidsentinel.associated_vnfc_cp_id
}


def get_instantiated_vnf_info():
    instantiated_vnf_info = {
        'flavour_id': uuidsentinel.flavour_id,
        'vnf_state': 'STARTED',
        'instance_id': uuidsentinel.instance_id
    }
    return instantiated_vnf_info


instantiated_vnf_info = {
    'ext_cp_info': [vnf_ext_cp_info],
    'flavour_id': uuidsentinel.flavour_id,
    'vnf_state': 'STARTED',
    'vnf_instance_id': uuidsentinel.vnf_instance_id
}


def vnf_resource_model_object(vnf_resource):
    resource_dict = {
        'id': vnf_resource.id,
        'vnf_instance_id': vnf_resource.vnf_instance_id,
        'resource_name': vnf_resource.resource_name,
        'resource_type': vnf_resource.resource_type,
        'resource_identifier': vnf_resource.resource_identifier,
        'resource_status': vnf_resource.resource_status
    }

    vnf_resource_db_obj = models.VnfResource()
    vnf_resource_db_obj.update(resource_dict)
    return vnf_resource_db_obj


def vnf_instance_model_object(vnf_instance):
    instance_dict = {
        'id': vnf_instance.id,
        'vnf_instance_name': vnf_instance.vnf_instance_name,
        'vnf_instance_description': vnf_instance.vnf_instance_description,
        'instantiation_state': vnf_instance.instantiation_state,
        'task_state': vnf_instance.task_state,
        'vnfd_id': vnf_instance.vnfd_id,
        'vnf_provider': vnf_instance.vnf_provider,
        'vnf_product_name': vnf_instance.vnf_product_name,
        'vnf_software_version': vnf_instance.vnf_software_version,
        'vnfd_version': vnf_instance.vnfd_version,
        'vim_connection_info': vnf_instance.vim_connection_info,
        'tenant_id': vnf_instance.tenant_id,
        'created_at': vnf_instance.created_at,
        'vnf_pkg_id': vnf_instance.vnf_pkg_id,
        'vnf_metadata': vnf_instance.vnf_metadata,
    }

    vnf_instance_db_obj = models.VnfInstance()
    vnf_instance_db_obj.update(instance_dict)
    return vnf_instance_db_obj


def get_changed_info_data():
    return {
        "vnf_instance_name": "",
        "vnf_instance_description": "",
        "metadata": {"test:": "test_value"},
        "vnf_configurable_properties": {"test": "test_value"},
        "vnfc_info_modifications_delete_ids": ["test1"],
        "vnfd_id": "2c69a161-0000-4b0f-bcf8-391f8fc76600",
        "vnf_provider": "NEC",
        "vnf_product_name": "MME",
        "vnf_software_version": "1.0",
        "vnfd_version": "MME_1.0"
    }


def get_vnf(vnfd_id, vim_id):
    return {
        'tenant_id': uuidsentinel.tenant_id,
        'name': "test_name",
        'vnfd_id': vnfd_id,
        'mgmt_ip_address': "test_mgmt_ip_address",
        'status': "ACTIVE",
        'description': "test_description",
        'placement_attr': "test_placement_attr",
        'vim_id': vim_id
    }


def get_changed_ext_conn_data():
    return [{
        "id": uuidsentinel.change_ext_conn_id,
        "resource_handle": {
            "vim_connection_id": uuidsentinel.vim_connection_id,
            "resource_id": uuidsentinel.vl_resource_id,
            "vim_level_resource_type": "OS::Neutron::Net",
        },
        "ext_link_ports": [{
            "id": uuidsentinel.ext_link_ports_id,
            "resource_handle": {
                "vim_connection_id": uuidsentinel.vim_connection_id,
                "resource_id": uuidsentinel.port_resource_id,
                "vim_level_resource_type": "OS::Neutron::Port",
            },
            "cp_instance_id": uuidsentinel.cp_instance_id,
        }]
    }]



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import base
from tacker.sol_refactored.objects import fields
from tacker.sol_refactored.objects.v2 import fields as v2fields


# NFV-SOL 005
# - v3.3.1 9.5.2.5 (API version: 2.1.0)
@base.TackerObjectRegistry.register
class VnfPkgInfoV2(base.TackerObject, base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vnfdId': fields.StringField(nullable=True),
        'vnfProvider': fields.StringField(nullable=True),
        'vnfProductName': fields.StringField(nullable=True),
        'vnfSoftwareVersion': fields.VersionField(nullable=True),
        'vnfdVersion': fields.VersionField(nullable=True),
        'compatibleSpecificationVersions': fields.ListOfVersionsField(
            nullable=True),
        'checksum': fields.ChecksumField(nullable=True),
        'packageSecurityOption': fields.EnumField(
            valid_values=[
                'OPTION_1',
                'OPTION_2',
            ],
            nullable=False,
        ),
        'signingCertificate': fields.StringField(nullable=True),
        'softwareImages': fields.ListOfObjectsField(
            'VnfPackageSoftwareImageInfoV2', nullable=True),
        'additionalArtifacts': fields.ListOfObjectsField(
            'VnfPackageArtifactInfoV2', nullable=True),
        'onboardingState': v2fields.PackageOnboardingStateTypeField(
            nullable=False),
        'operationalState': v2fields.PackageOperationalStateTypeField(
            nullable=False),
        'usageState': v2fields.PackageUsageStateTypeField(nullable=False),
        'vnfmInfo': fields.ListOfStringsField(nullable=False),
        'userDefinedData': fields.KeyValuePairsField(nullable=True),
        'onboardingFailureDetails': fields.ObjectField(
            'ProblemDetails', nullable=True),
        '_links': fields.ObjectField('VnfPkgInfoV2_Links', nullable=False),
    }


@base.TackerObjectRegistry.register
class VnfPkgInfoV2_Links(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'self': fields.ObjectField('Link', nullable=False),
        'vnfd': fields.ObjectField('Link', nullable=False),
        'packageContent': fields.ObjectField('Link', nullable=False),
    }



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import vnfd_utils
from tacker.tests import base


SAMPLE_VNFD_ID = "b1bb0ce7-ebca-4fa7-95ed-4840d7000000"
SAMPLE_FLAVOUR_ID = "simple"


class TestVnfd(base.BaseTestCase):

    def setUp(self):
        super(TestVnfd, self).setUp()
        cur_dir = os.path.dirname(__file__)
        sample_dir = os.path.join(cur_dir, "..", "samples")

        self.vnfd_1 = vnfd_utils.Vnfd(SAMPLE_VNFD_ID)
        self.vnfd_1.init_from_csar_dir(os.path.join(sample_dir, "sample1"))

    def test_get_sw_image(self):
        expected_result = {
            'VDU2': 'VDU2-image',
            'VirtualStorage': 'image-1.0.0-x86_64-disk'
        }
        result = self.vnfd_1.get_sw_image(SAMPLE_FLAVOUR_ID)

        self.assertEqual(expected_result, result)

    def test_get_sw_image_data(self):
        result = self.vnfd_1.get_sw_image_data(SAMPLE_FLAVOUR_ID)
        # check 'file' existence if 'artifact' is defined
        self.assertEqual('../Files/images/image-1.0.0-x86_64-disk.img',
                         result['VDU2']['file'])
        self.assertNotIn('file', result['VirtualStorage'])

    def test_get_nodes(self):
        expected_vdus = ['VDU1', 'VDU2']
        expected_storages = ['VirtualStorage']
        expected_vls = ['internalVL1', 'internalVL2', 'internalVL3']
        expected_cps = ['VDU1_CP1', 'VDU1_CP2', 'VDU1_CP3', 'VDU1_CP4',
                        'VDU1_CP5', 'VDU2_CP1', 'VDU2_CP2', 'VDU2_CP3',
                        'VDU2_CP4', 'VDU2_CP5']
        # check keys and sampling data
        result_vdus = self.vnfd_1.get_vdu_nodes(SAMPLE_FLAVOUR_ID)
        self.assertEqual(expected_vdus, list(result_vdus.keys()))
        self.assertEqual('VDU1 compute node',
            result_vdus['VDU1']['properties']['description'])

        result_storages = self.vnfd_1.get_storage_nodes(SAMPLE_FLAVOUR_ID)
        self.assertEqual(expected_storages, list(result_storages.keys()))
        self.assertEqual('1.0.0',
            result_storages['VirtualStorage']['properties']['sw_image_data']
            ['version'])

        result_vls = self.vnfd_1.get_virtual_link_nodes(SAMPLE_FLAVOUR_ID)
        self.assertEqual(expected_vls, list(result_vls.keys()))
        self.assertEqual(['ipv4'],
            result_vls['internalVL3']['properties']['connectivity_type']
            ['layer_protocols'])

        result_cps = self.vnfd_1.get_vducp_nodes(SAMPLE_FLAVOUR_ID)
        self.assertEqual(expected_cps, list(result_cps.keys()))
        self.assertEqual(0,
            result_cps['VDU2_CP1']['properties']['order'])

    def test_get_vdu_cps(self):
        expected_result = ['VDU1_CP1', 'VDU1_CP2', 'VDU1_CP3',
                           'VDU1_CP4', 'VDU1_CP5']
        result = self.vnfd_1.get_vdu_cps(SAMPLE_FLAVOUR_ID, 'VDU1')

        self.assertEqual(expected_result, result)

    def test_get_vdu_storages(self):
        vdu_nodes = self.vnfd_1.get_vdu_nodes(SAMPLE_FLAVOUR_ID)
        expected_result = ['VirtualStorage']
        result = self.vnfd_1.get_vdu_storages(vdu_nodes['VDU1'])
        self.assertEqual(expected_result, result)

        expected_result = []
        result = self.vnfd_1.get_vdu_storages(vdu_nodes['VDU2'])
        self.assertEqual(expected_result, result)

    def test_get_base_hot(self):
        result = self.vnfd_1.get_base_hot(SAMPLE_FLAVOUR_ID)
        # check keys and sampling data
        self.assertEqual(['VDU1.yaml'], list(result['files'].keys()))
        self.assertEqual({'get_param': 'net3'},
            result['files']['VDU1.yaml']['resources']['VDU1_CP3']
            ['properties']['network'])

    def test_get_vl_name_from_cp(self):
        vdu_cps = self.vnfd_1.get_vducp_nodes(SAMPLE_FLAVOUR_ID)
        result = self.vnfd_1.get_vl_name_from_cp(SAMPLE_FLAVOUR_ID,
            vdu_cps['VDU1_CP1'])
        # externalVL
        self.assertEqual(None, result)

        result = self.vnfd_1.get_vl_name_from_cp(SAMPLE_FLAVOUR_ID,
            vdu_cps['VDU1_CP3'])
        self.assertEqual('internalVL1', result)

    def test_get_compute_flavor(self):
        result = self.vnfd_1.get_compute_flavor(SAMPLE_FLAVOUR_ID, 'VDU1')
        self.assertEqual('m1.tiny', result)

    def test_get_default_instantiation_level(self):
        result = self.vnfd_1.get_default_instantiation_level(SAMPLE_FLAVOUR_ID)
        self.assertEqual('instantiation_level_1', result)

    def test_get_vdu_num(self):
        result = self.vnfd_1.get_vdu_num(SAMPLE_FLAVOUR_ID, 'VDU1',
            'instantiation_level_2')
        self.assertEqual(3, result)

    def test_get_placement_groups(self):
        expected_result = {'affinityOrAntiAffinityGroup1': ['VDU1', 'VDU2']}
        result = self.vnfd_1.get_placement_groups(SAMPLE_FLAVOUR_ID)
        self.assertEqual(expected_result, result)

    def test_get_tartget(self):
        result = self.vnfd_1.get_affinity_targets(SAMPLE_FLAVOUR_ID)
        self.assertEqual([], result)

        expected_result = [(['VDU1', 'VDU2'], 'nfvi_node')]
        result = self.vnfd_1.get_anti_affinity_targets(SAMPLE_FLAVOUR_ID)
        self.assertEqual(expected_result, result)

    def test_get_interface_script(self):
        # script specified
        result = self.vnfd_1.get_interface_script(SAMPLE_FLAVOUR_ID,
            "instantiate_start")
        self.assertEqual("../Scripts/sample_script.py", result)

        # [] specified
        result = self.vnfd_1.get_interface_script(SAMPLE_FLAVOUR_ID,
            "scale_start")
        self.assertEqual(None, result)

        # not specified
        result = self.vnfd_1.get_interface_script(SAMPLE_FLAVOUR_ID,
            "scale_end")
        self.assertEqual(None, result)

    def test_get_scale_vdu_and_num(self):
        expected_result = {'VDU1': 1}
        result = self.vnfd_1.get_scale_vdu_and_num(SAMPLE_FLAVOUR_ID,
            'VDU1_scale')
        self.assertEqual(expected_result, result)

    def test_get_scale_vdu_and_num_no_delta(self):
        self.assertRaises(sol_ex.DeltaMissingInVnfd,
            self.vnfd_1.get_scale_vdu_and_num, SAMPLE_FLAVOUR_ID,
            'Invalid_scale')

    def test_get_scale_info_from_inst_level(self):
        expected_result = {'VDU1_scale': {'scale_level': 2}}
        result = self.vnfd_1.get_scale_info_from_inst_level(
            SAMPLE_FLAVOUR_ID, 'instantiation_level_2')
        self.assertEqual(expected_result, result)

    def test_get_max_scale_level(self):
        result = self.vnfd_1.get_max_scale_level(SAMPLE_FLAVOUR_ID,
            'VDU1_scale')
        self.assertEqual(2, result)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_messaging
from tacker.common import rpc
from tacker.common import topics
from tacker.objects import base as objects_base


class VNFLcmRPCAPI(object):

    target = oslo_messaging.Target(
        exchange='tacker',
        topic=topics.TOPIC_CONDUCTOR,
        fanout=False,
        version='1.0')

    def instantiate(self, context, vnf_instance, vnf,
                    instantiate_vnf, vnf_lcm_op_occs_id,
                    cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'instantiate',
                          vnf_instance=vnf_instance,
                          vnf_dict=vnf,
                          instantiate_vnf=instantiate_vnf,
                          vnf_lcm_op_occs_id=vnf_lcm_op_occs_id)

    def terminate(self, context, vnf_instance, vnf,
                terminate_vnf_req, vnf_lcm_op_occs_id,
                cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'terminate',
                          vnf_lcm_op_occs_id=vnf_lcm_op_occs_id,
                          vnf_instance=vnf_instance,
                          terminate_vnf_req=terminate_vnf_req,
                          vnf_dict=vnf)

    def heal(self, context, vnf_instance, vnf_dict, heal_vnf_request,
             vnf_lcm_op_occs_id, cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'heal',
                          vnf_instance=vnf_instance,
                          vnf_dict=vnf_dict,
                          heal_vnf_request=heal_vnf_request,
                          vnf_lcm_op_occs_id=vnf_lcm_op_occs_id)

    def update(self, context, vnf_lcm_opoccs, body_data,
            vnfd_pkg_data, vnfd_id, cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'update',
                          vnf_lcm_opoccs=vnf_lcm_opoccs,
                          body_data=body_data,
                          vnfd_pkg_data=vnfd_pkg_data,
                          vnfd_id=vnfd_id)

    def scale(
            self,
            context,
            vnf_info,
            vnf_instance,
            scale_vnf_request,
            cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'scale',
                          vnf_info=vnf_info,
                          vnf_instance=vnf_instance,
                          scale_vnf_request=scale_vnf_request)

    def send_notification(self, context, notification, cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'send_notification',
                          notification=notification)

    def test_notification(self, context,
            vnf_lcm_subscription=None, cast=False):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'test_notification',
                          vnf_lcm_subscription=vnf_lcm_subscription)

    def rollback(self, context, vnf_info, vnf_instance,
            operation_params, cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(context, 'rollback',
                          vnf_info=vnf_info,
                          vnf_instance=vnf_instance,
                          operation_params=operation_params)

    def change_ext_conn(
            self,
            context,
            vnf_instance,
            vnf_dict,
            change_ext_conn_req,
            vnf_lcm_op_occs_id,
            cast=True):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        rpc_method = cctxt.cast if cast else cctxt.call
        return rpc_method(
            context,
            'change_ext_conn',
            vnf_instance=vnf_instance,
            vnf_dict=vnf_dict,
            change_ext_conn_req=change_ext_conn_req,
            vnf_lcm_op_occs_id=vnf_lcm_op_occs_id)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import models
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy import DateTime
from sqlalchemy import orm

from tacker.db import model_base
from tacker.db import models_v1
from tacker.db import types


class VnfPackageUserData(model_base.BASE, models.SoftDeleteMixin,
                         models.TimestampMixin):
    """Contains all info about vnf packages UserDefinedData."""

    __tablename__ = 'vnf_packages_user_data'
    id = sa.Column(sa.Integer, nullable=False, primary_key=True)
    package_uuid = sa.Column(sa.String(36),
                             sa.ForeignKey('vnf_packages.id'),
                             nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.String(255), nullable=False)


class VnfSoftwareImageMetadata(model_base.BASE, models.SoftDeleteMixin,
                               models.TimestampMixin):
    """Contains all info about vnf packages software image metadata."""

    __tablename__ = 'vnf_software_image_metadata'
    id = sa.Column(sa.Integer, nullable=False, primary_key=True)
    image_uuid = sa.Column(sa.String(36),
                           sa.ForeignKey('vnf_software_images.id'),
                           nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.String(255), nullable=False)


class VnfSoftwareImage(model_base.BASE, models.SoftDeleteMixin,
                       models.TimestampMixin, models_v1.HasId):
    """Contains all info about vnf packages software images."""

    __tablename__ = 'vnf_software_images'
    software_image_id = sa.Column(sa.Integer, nullable=False)
    flavour_uuid = sa.Column(sa.String(36), sa.ForeignKey(
        'vnf_deployment_flavours.id'), nullable=False)
    name = sa.Column(sa.String(255), nullable=True)
    provider = sa.Column(sa.String(255), nullable=True)
    version = sa.Column(sa.String(255), nullable=True)
    algorithm = sa.Column(sa.String(64), nullable=True)
    hash = sa.Column(sa.String(128), nullable=True)
    container_format = sa.Column(sa.String(20), nullable=True)
    disk_format = sa.Column(sa.String(20), nullable=True)
    min_disk = sa.Column(sa.Integer, nullable=False)
    min_ram = sa.Column(sa.Integer, nullable=False)
    size = sa.Column(sa.BigInteger, nullable=False)
    image_path = sa.Column(sa.Text, nullable=False)

    _metadata = orm.relationship(
        VnfSoftwareImageMetadata,
        primaryjoin='and_(VnfSoftwareImage.id == '
                    'VnfSoftwareImageMetadata.image_uuid)')

    @property
    def metadetails(self):
        return {m.key: m.value for m in self._metadata}


class VnfArtifactMetadata(model_base.BASE, models.SoftDeleteMixin,
                          models.TimestampMixin):
    """Contains all info about vnf packages artifacts metadata."""

    __tablename__ = 'vnf_artifact_metadata'
    id = sa.Column(sa.Integer, nullable=False, primary_key=True)
    artifact_uuid = sa.Column(sa.String(36),
                              sa.ForeignKey('vnf_artifacts.id'),
                              nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.String(255), nullable=False)


class VnfDeploymentFlavour(model_base.BASE, models.SoftDeleteMixin,
                models.TimestampMixin, models_v1.HasId):
    """Contains all info about vnf packages Deployment Flavours."""

    __tablename__ = 'vnf_deployment_flavours'
    package_uuid = sa.Column(sa.String(36),
                             sa.ForeignKey('vnf_packages.id'),
                             nullable=False)
    flavour_id = sa.Column(sa.String(255), nullable=False)
    flavour_description = sa.Column(sa.Text(), nullable=False)
    instantiation_levels = sa.Column(sa.Text(), nullable=True)

    software_images = orm.relationship(
        VnfSoftwareImage, primaryjoin='and_(VnfDeploymentFlavour.id == '
                                      'VnfSoftwareImage.flavour_uuid,'
                                      'VnfSoftwareImage.deleted == 0)')


class VnfPackageVnfdSoftDeleteMixin(object):
    deleted_at = sa.Column(DateTime)
    deleted = sa.Column(sa.String(36), default='0')

    def soft_delete(self, session):
        """Mark this object as deleted."""
        self.deleted = self.id
        self.deleted_at = timeutils.utcnow()
        self.save(session=session)


class VnfPackageVnfd(model_base.BASE, VnfPackageVnfdSoftDeleteMixin,
                  models.TimestampMixin, models_v1.HasId):
    """Contains all info about vnf packages VNFD."""

    __tablename__ = 'vnf_package_vnfd'
    __table_args__ = (
        sa.schema.UniqueConstraint(
            "vnfd_id",
            "deleted",
            name="uniq_vnf_package_vnfd0vnfd_id0deleted"),
    )

    package_uuid = sa.Column(sa.String(36),
                             sa.ForeignKey('vnf_packages.id'),
                             nullable=False)
    vnfd_id = sa.Column(types.Uuid, nullable=False)
    vnf_provider = sa.Column(sa.String(255), nullable=False)
    vnf_product_name = sa.Column(sa.String(255), nullable=False)
    vnf_software_version = sa.Column(sa.String(255), nullable=False)
    vnfd_version = sa.Column(sa.String(255), nullable=False)


class VnfPackageArtifactInfo(model_base.BASE, models.SoftDeleteMixin,
                 models.TimestampMixin, models_v1.HasId):
    """Contains all info about vnf artifacts."""

    __tablename__ = 'vnf_artifacts'
    package_uuid = sa.Column(sa.String(36),
                             sa.ForeignKey('vnf_packages.id'),
                             nullable=False)
    artifact_path = sa.Column(sa.Text(), nullable=False)
    algorithm = sa.Column(sa.String(64), nullable=False)
    hash = sa.Column(sa.String(128), nullable=False)
    _metadata = sa.Column(sa.JSON(), nullable=True)


class VnfPackage(model_base.BASE, models.SoftDeleteMixin,
                 models.TimestampMixin, models_v1.HasTenant,
                 models_v1.HasId):
    """Contains all info about vnf packages."""

    __tablename__ = 'vnf_packages'
    onboarding_state = sa.Column(sa.String(255), nullable=False)
    operational_state = sa.Column(sa.String(255), nullable=False)
    usage_state = sa.Column(sa.String(255), nullable=False)
    algorithm = sa.Column(sa.String(64), nullable=True)
    hash = sa.Column(sa.String(128), nullable=True)
    location_glance_store = sa.Column(sa.Text(), nullable=True)
    size = sa.Column(sa.BigInteger, nullable=False, default=0)

    _metadata = orm.relationship(
        VnfPackageUserData,
        primaryjoin='and_(VnfPackage.id == '
                    'VnfPackageUserData.package_uuid)')

    vnf_deployment_flavours = orm.relationship(
        VnfDeploymentFlavour,
        primaryjoin='and_(VnfPackage.id == '
                    'VnfDeploymentFlavour.package_uuid,'
                    'VnfDeploymentFlavour.deleted == 0)')

    vnfd = orm.relationship(
        VnfPackageVnfd, uselist=False,
        primaryjoin='and_(VnfPackage.id == '
                    'VnfPackageVnfd.package_uuid,'
                    'VnfPackageVnfd.deleted == 0)')

    vnf_artifacts = orm.relationship(
        VnfPackageArtifactInfo,
        primaryjoin='and_(VnfPackage.id == '
                    'VnfPackageArtifactInfo.package_uuid,'
                    'VnfPackageArtifactInfo.deleted == 0)')

    @property
    def metadetails(self):
        return {m.key: m.value for m in self._metadata}


class VnfInstance(model_base.BASE, models.SoftDeleteMixin,
                models.TimestampMixin, models_v1.HasId):
    """Represents a Vnf Instance."""

    __tablename__ = 'vnf_instances'
    vnf_instance_name = sa.Column(sa.String(255), nullable=True)
    vnf_instance_description = sa.Column(sa.String(1024), nullable=True)
    vnf_provider = sa.Column(sa.String(255), nullable=False)
    vnf_product_name = sa.Column(sa.String(255), nullable=False)
    vnf_software_version = sa.Column(sa.String(255), nullable=False)
    vnfd_version = sa.Column(sa.String(255), nullable=False)
    vnfd_id = sa.Column(types.Uuid, nullable=False)
    instantiation_state = sa.Column(sa.String(255), nullable=False)
    task_state = sa.Column(sa.String(255), nullable=True)
    vim_connection_info = sa.Column(sa.JSON(), nullable=True)
    tenant_id = sa.Column('tenant_id', sa.String(length=64), nullable=False)
    vnf_pkg_id = sa.Column(types.Uuid, nullable=False)
    vnf_metadata = sa.Column(sa.JSON(), nullable=True)


class VnfInstantiatedInfo(model_base.BASE, models.SoftDeleteMixin,
                    models.TimestampMixin):
    """Contain the details of VNF instance"""

    __tablename__ = 'vnf_instantiated_info'
    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    vnf_instance_id = sa.Column(sa.String,
                            sa.ForeignKey('vnf_instances.id'),
                            nullable=False)
    flavour_id = sa.Column(sa.String(255), nullable=False)
    scale_status = sa.Column(sa.JSON(), nullable=True)
    ext_cp_info = sa.Column(sa.JSON(), nullable=False)
    ext_virtual_link_info = sa.Column(sa.JSON(), nullable=True)
    ext_managed_virtual_link_info = sa.Column(sa.JSON(), nullable=True)
    vnfc_resource_info = sa.Column(sa.JSON(), nullable=True)
    vnf_virtual_link_resource_info = sa.Column(sa.JSON(), nullable=True)
    virtual_storage_resource_info = sa.Column(sa.JSON(), nullable=True)
    vnfc_info = sa.Column(sa.JSON(), nullable=True)
    vnf_state = sa.Column(sa.String(255), nullable=False)
    instance_id = sa.Column(sa.Text(), nullable=True)
    instantiation_level_id = sa.Column(sa.String(255), nullable=True)
    additional_params = sa.Column(sa.JSON(), nullable=True)

    vnf_instance = orm.relationship(
        VnfInstance,
        backref=orm.backref(
            'instantiated_vnf_info',
            uselist=False),
        foreign_keys=vnf_instance_id,
        primaryjoin='and_(VnfInstantiatedInfo.vnf_instance_id == '
        'VnfInstance.id, VnfInstantiatedInfo.deleted == 0)')


class VnfResource(model_base.BASE, models.SoftDeleteMixin,
                models.TimestampMixin, models_v1.HasId):
    """Resources belongs to the VNF"""

    __tablename__ = 'vnf_resources'
    vnf_instance_id = sa.Column(sa.String(36),
                                sa.ForeignKey('vnf_instances.id'),
                                nullable=False)
    resource_name = sa.Column(sa.Text(), nullable=True)
    resource_type = sa.Column(sa.String(255), nullable=False)
    resource_identifier = sa.Column(sa.String(255), nullable=False)
    resource_status = sa.Column(sa.String(255), nullable=False)
    tenant_id = sa.Column('tenant_id', sa.String(length=64), nullable=False)


class VnfLcmFilters(model_base.BASE):
    """Contains all info about vnf LCM filters."""

    __tablename__ = 'vnf_lcm_filters'
    __maxsize__ = 65536
    id = sa.Column(sa.Integer, nullable=True, primary_key=True)
    subscription_uuid = sa.Column(sa.String(36),
                            sa.ForeignKey('vnf_lcm_subscriptions.id'),
                            nullable=False)
    filter = sa.Column(sa.JSON, nullable=False)
    vnf_products_from_providers = sa.Column(sa.JSON, nullable=True)
    vnfd_ids = sa.Column(sa.Text(), nullable=True)
    vnfd_ids_len = sa.Column(sa.Integer, nullable=True)
    vnf_provider = sa.Column(sa.Text(), nullable=True)
    vnf_product_name = sa.Column(sa.Text(), nullable=True)
    vnf_software_version = sa.Column(sa.Text(), nullable=True)
    vnfd_versions = sa.Column(sa.Text(), nullable=True)
    vnfd_versions_len = sa.Column(sa.Integer, nullable=True)
    vnf_instance_ids = sa.Column(sa.Text(), nullable=True)
    vnf_instance_ids_len = sa.Column(sa.Integer, nullable=True)
    vnf_instance_names = sa.Column(sa.Text(), nullable=True)
    vnf_instance_names_len = sa.Column(sa.Integer, nullable=True)
    notification_types = sa.Column(
        sa.LargeBinary(
            length=__maxsize__),
        nullable=True)
    notification_types_len = sa.Column(sa.Integer, nullable=True)
    operation_types = sa.Column(
        sa.LargeBinary(
            length=__maxsize__),
        nullable=True)
    operation_types_len = sa.Column(sa.Integer, nullable=True)
    operation_states = sa.Column(sa.Text(), nullable=True)
    operation_states_len = sa.Column(sa.Integer, nullable=True)


class VnfLcmSubscriptions(model_base.BASE, models.SoftDeleteMixin,
                models.TimestampMixin):
    """Contains all info about vnf LCM Subscriptions."""

    __tablename__ = 'vnf_lcm_subscriptions'
    id = sa.Column(sa.String(36), nullable=False, primary_key=True)
    callback_uri = sa.Column(sa.String(255), nullable=False)
    authentication = sa.Column(sa.JSON, nullable=True)
    tenant_id = sa.Column('tenant_id', sa.String(length=64), nullable=False)

    subscription_filter = orm.relationship(
        VnfLcmFilters,
        primaryjoin='and_(VnfLcmSubscriptions.id == '
                    'VnfLcmFilters.subscription_uuid)')


class VnfLcmOpOccs(model_base.BASE, models.SoftDeleteMixin,
                models.TimestampMixin):
    """VNF LCM OP OCCS Fields"""

    __tablename__ = 'vnf_lcm_op_occs'
    id = sa.Column(sa.String(16), primary_key=True)
    vnf_instance_id = sa.Column(sa.String(36),
                                sa.ForeignKey('vnf_instances.id'),
                                nullable=False)
    grant_id = sa.Column(sa.String(36), nullable=True)
    state_entered_time = sa.Column(sa.DateTime(), nullable=False)
    start_time = sa.Column(sa.DateTime(), nullable=False)
    operation_state = sa.Column(sa.String(length=255), nullable=False)
    operation = sa.Column(sa.String(length=255), nullable=False)
    is_automatic_invocation = sa.Column(sa.Boolean, nullable=False)
    operation_params = sa.Column(sa.JSON(), nullable=True)
    is_cancel_pending = sa.Column(sa.Boolean(), nullable=False)
    error = sa.Column(sa.JSON(), nullable=True)
    resource_changes = sa.Column(sa.JSON(), nullable=True)
    changed_info = sa.Column(sa.JSON(), nullable=True)
    changed_ext_connectivity = sa.Column(sa.JSON(), nullable=True)
    error_point = sa.Column(sa.Integer, nullable=False)
    tenant_id = sa.Column('tenant_id', sa.String(length=64), nullable=False)


class PlacementConstraint(model_base.BASE, models.SoftDeleteMixin,
                models.TimestampMixin, models_v1.HasId):
    """Represents a Vnf Placement Constraint."""

    __tablename__ = 'placement_constraint'
    vnf_instance_id = sa.Column(sa.String(36),
                                sa.ForeignKey('vnf_instances.id'),
                                nullable=False)
    affinity_or_anti_affinity = sa.Column(sa.String(255), nullable=False)
    scope = sa.Column(sa.String(255), nullable=False)
    server_group_name = sa.Column(sa.String(255), nullable=False)
    resource = sa.Column(sa.JSON(), nullable=True)
